__all__ = [
    "FromDishka",
    "inject",
    "setup_dishka",
]

from collections.abc import Iterable
from typing import Any, cast

from sanic import HTTPResponse, Request, Sanic
from sanic.models.handler_types import RouteHandler
from sanic_routing import Route

from dishka import AsyncContainer, FromDishka
from dishka.integrations.base import is_dishka_injected, wrap_injection


def inject(func: RouteHandler) -> RouteHandler:
    return cast(
        RouteHandler,
        wrap_injection(
            func=func,
            is_async=True,
            container_getter=lambda args, _: args[0].ctx.dishka_container,
        ),
    )


class ContainerMiddleware:
    def __init__(self, container: AsyncContainer) -> None:
        self.container = container

    async def on_request(self, request: Request) -> None:
        request.ctx.container_wrapper = self.container({Request: request})
        request.ctx.dishka_container = await request.ctx.container_wrapper.__aenter__()  # noqa: E501

    async def on_response(self, request: Request, _: HTTPResponse) -> None:
        await request.ctx.dishka_container.close()


def _inject_routes(routes: Iterable[Route]) -> None:
    for route in routes:
        if not is_dishka_injected(route.handler):
            route.handler = inject(route.handler)


def setup_dishka(
    container: AsyncContainer,
    app: Sanic[Any, Any],
    *,
    auto_inject: bool = False,
) -> None:
    middleware = ContainerMiddleware(container)
    app.on_request(middleware.on_request)
    app.on_response(middleware.on_response)  # type: ignore[no-untyped-call]

    if auto_inject:
        _inject_routes(app.router.routes)
        for blueprint in app.blueprints.values():
            _inject_routes(blueprint.routes)



from dataclasses import dataclass
from enum import Enum
from typing import Any


@dataclass(slots=True)
class _ScopeValue:
    name: str
    skip: bool


def new_scope(value: str, *, skip: bool = False) -> _ScopeValue:
    return _ScopeValue(value, skip)


class BaseScope(Enum):
    __slots__ = ("name", "skip")

    def __init__(self, value: _ScopeValue) -> None:
        self.name = value.name  # type: ignore[misc]
        self.skip = value.skip


class Scope(BaseScope):
    RUNTIME = new_scope("RUNTIME", skip=True)
    APP = new_scope("APP")
    SESSION = new_scope("SESSION", skip=True)
    REQUEST = new_scope("REQUEST")
    ACTION = new_scope("ACTION")
    STEP = new_scope("STEP")


class InvalidScopes(BaseScope):
    UNKNOWN_SCOPE = new_scope("<unknown scope>", skip=True)

    def __str__(self) -> Any:
        return self.value



__all__ = [
    "FromDishka",
    "inject",
    "setup_dishka",
]

from collections.abc import Callable
from typing import ParamSpec, TypeVar

from starlette.applications import Starlette
from starlette.requests import Request
from starlette.types import ASGIApp, Receive, Scope, Send
from starlette.websockets import WebSocket

from dishka import AsyncContainer, FromDishka
from dishka import Scope as DIScope
from .base import wrap_injection

T = TypeVar("T")
P = ParamSpec("P")

def inject(func: Callable[P, T]) -> Callable[P, T]:
    return wrap_injection(
        func=func,
        is_async=True,
        container_getter=lambda r, _: r[0].scope["state"]["dishka_container"],
    )


class ContainerMiddleware:
    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(
            self, scope: Scope, receive: Receive, send: Send,
    ) -> None:
        if scope["type"] not in ("http", "websocket"):
            return await self.app(scope, receive, send)

        request: Request | WebSocket
        context: dict[type[Request | WebSocket], Request | WebSocket]

        if scope["type"] == "http":
            request = Request(scope)
            context = {Request: request}
            di_scope = DIScope.REQUEST

        else:
            request = WebSocket(scope, receive, send)
            context = {WebSocket: request}
            di_scope = DIScope.SESSION

        async with request.app.state.dishka_container(
            context, scope=di_scope,
        ) as request_container:
            request.state.dishka_container = request_container
            return await self.app(scope, receive, send)


def setup_dishka(container: AsyncContainer, app: Starlette) -> None:
    app.add_middleware(ContainerMiddleware)
    app.state.dishka_container = container



from collections.abc import Callable
from typing import Any, overload

from .composite import CompositeDependencySource, ensure_composite
from .decorator import Decorator
from .make_factory import make_factory
from .unpack_provides import unpack_decorator


def _decorate(
        source: Callable[..., Any] | type,
        provides: Any = None,
        *,
        is_in_class: bool = True,
) -> CompositeDependencySource:
    composite = ensure_composite(source)
    decorator = Decorator(
        make_factory(
            provides=provides,
            scope=None,
            source=source,
            cache=False,
            is_in_class=is_in_class,
        ),
    )
    if (
        decorator.provides not in decorator.factory.kw_dependencies.values()
        and decorator.provides not in decorator.factory.dependencies
    ):
        name = getattr(source, "__qualname__", "") or str(source)
        raise ValueError(
            f"Decorator {name} does not depends on provided type.\n"
            f"Did you mean @provide instead of @decorate?",
        )

    composite.dependency_sources.extend(unpack_decorator(decorator))
    return composite


@overload
def decorate(
        *,
        provides: Any = None,
) -> Callable[
    [Callable[..., Any]], CompositeDependencySource,
]:
    ...


@overload
def decorate(
        source: Callable[..., Any] | type,
        *,
        provides: Any = None,
) -> CompositeDependencySource:
    ...


def decorate(
        source: Callable[..., Any] | type | None = None,
        provides: Any = None,
) -> CompositeDependencySource | Callable[
    [Callable[..., Any]], CompositeDependencySource,
]:
    if source is not None:
        return _decorate(source, provides, is_in_class=True)

    def scoped(func: Callable[..., Any]) -> CompositeDependencySource:
        return _decorate(func, provides, is_in_class=True)

    return scoped


def decorate_on_instance(
        source: Callable[..., Any] | type,
        provides: Any = None,
) -> CompositeDependencySource:
    return _decorate(source, provides, is_in_class=False)




from collections.abc import Iterator, MutableMapping
from typing import Any, NoReturn

from .entities.key import DependencyKey


class ContextProxy(MutableMapping[DependencyKey, Any]):
    def __init__(
            self,
            context: dict[DependencyKey, Any],
            cache: dict[DependencyKey, Any],
    ) -> None:
        self._cache = cache
        self._context = context

    def __setitem__(self, key: DependencyKey, value: Any) -> None:
        self._cache[key] = value
        self._context[key] = value

    def __delitem__(self, key: DependencyKey) -> NoReturn:
        raise RuntimeError("Cannot delete anything from context")

    def __getitem__(self, key: DependencyKey) -> Any:
        return self._cache[key]

    def __len__(self) -> int:
        return len(self._cache)

    def __iter__(self) -> Iterator[DependencyKey]:
        return iter(self._cache)



from collections.abc import Sequence

from dishka.entities.key import DependencyKey
from dishka.entities.provides_marker import ProvideMultiple
from .alias import Alias
from .composite import DependencySource
from .decorator import Decorator
from .factory import Factory


def unpack_factory(factory: Factory) -> Sequence[DependencySource]:
    if not isinstance(factory.provides.type_hint, ProvideMultiple):
        return [factory]

    provides_first, *provides_others = factory.provides.type_hint.items

    res: list[DependencySource] = [
        Alias(
            provides=DependencyKey(provides_other, factory.provides.component),
            source=DependencyKey(provides_first, factory.provides.component),
            cache=factory.cache,
        )
        for provides_other in provides_others
    ]
    res.append(
        Factory(
            dependencies=factory.dependencies,
            kw_dependencies=factory.kw_dependencies,
            type_=factory.type,
            source=factory.source,
            scope=factory.scope,
            is_to_bind=factory.is_to_bind,
            cache=factory.cache,
            provides=DependencyKey(
                provides_first,
                factory.provides.component,
            ),
        ),
    )
    return res


def unpack_decorator(decorator: Decorator) -> Sequence[DependencySource]:
    if not isinstance(decorator.provides.type_hint, ProvideMultiple):
        return [decorator]

    return [
        Decorator(
            factory=decorator.factory,
            provides=DependencyKey(provides, decorator.provides.component),
        )
        for provides in decorator.provides.type_hint.items
    ]


def unpack_alias(alias: Alias) -> Sequence[DependencySource]:
    if not isinstance(alias.provides.type_hint, ProvideMultiple):
        return [alias]

    return [
        Alias(
            provides=DependencyKey(provides, alias.provides.component),
            source=alias.source,
            cache=alias.cache,
        )
        for provides in alias.provides.type_hint.items
    ]



from __future__ import annotations

import inspect
from collections.abc import Callable, Sequence
from typing import Any, TypeGuard

from dishka.entities.component import DEFAULT_COMPONENT, Component
from dishka.entities.scope import BaseScope
from .dependency_source import (
    Alias,
    ContextVariable,
    Decorator,
    DependencySource,
    Factory,
    alias,
    from_context,
)
from .dependency_source.composite import CompositeDependencySource
from .dependency_source.make_decorator import decorate_on_instance
from .dependency_source.make_factory import (
    provide_all_on_instance,
    provide_on_instance,
)


def is_dependency_source(
    attribute: Any,
) -> TypeGuard[CompositeDependencySource]:
    return isinstance(attribute, CompositeDependencySource)


class BaseProvider:
    def __init__(self, component: Component | None) -> None:
        if component is not None:
            self.component = component
        self.factories: list[Factory] = []
        self.aliases: list[Alias] = []
        self.decorators: list[Decorator] = []
        self.context_vars: list[ContextVariable] = []


class Provider(BaseProvider):
    """
    A collection of dependency sources.

    Inherit this class and add attributes using
    `provide`, `alias` or `decorate`.

    You can use `__init__`, regular methods and attributes as usual,
    they won't be analyzed when creating a container

    The only intended usage of providers is to pass them when
    creating a container
    """
    scope: BaseScope | None = None
    component: Component = DEFAULT_COMPONENT

    def __init__(
            self,
            scope: BaseScope | None = None,
            component: Component | None = None,
    ):
        super().__init__(component)
        self.scope = self.scope or scope
        self._init_dependency_sources()

    def _init_dependency_sources(self) -> None:
        sources = inspect.getmembers(self, is_dependency_source)
        sources.sort(key=lambda s: s[1].number)
        for name, composite in sources:
            self._add_dependency_sources(name, composite.dependency_sources)

    def _name(self) -> str:
        if type(self) is Provider:
            return str(self)
        else:
            cls = type(self)
            return f"`{cls.__module__}.{cls.__qualname__}`"

    def _source_name(self, factory: Factory) -> str:
        source = factory.source
        if source == factory.provides.type_hint:
            return "`provides()`"
        elif func := getattr(source, "__func__", None):
            name = getattr(func, "__qualname__", None)
            if name:
                return f"`{name}`"
        elif isinstance(source, type):
            name = getattr(source, "__qualname__", None)
            if name:
                return f"`{source.__module__}.{name}`"
        else:
            name = getattr(source, "__qualname__", None)
            if name:
                return f"`{name}`"
        return str(source)

    def _provides_name(self, factory: Factory | ContextVariable) -> str:
        hint = factory.provides.type_hint
        name = getattr(hint, "__qualname__", None)
        if name:
            return f"`{hint.__module__}.{name}`"
        return str(hint)

    def _add_dependency_sources(
            self, name: str, sources: Sequence[DependencySource],
    ) -> None:
        for source in sources:
            if isinstance(source, Alias):
                self.aliases.append(source)
            if isinstance(source, Factory):
                if source.scope is None:
                    src_name = self._source_name(source)
                    provides_name = self._provides_name(source)
                    raise ValueError(
                        f"No scope is set for {provides_name}.\n"
                        f"Set in provide() call for {src_name} or "
                        f"within {self._name()}",
                    )
                self.factories.append(source)
            if isinstance(source, Decorator):
                self.decorators.append(source)
            if isinstance(source, ContextVariable):
                if source.scope is None:
                    provides_name = self._provides_name(source)
                    raise ValueError(
                        f"No scope is set for {provides_name}.\n"
                        f"Set in from_context() call or within {self._name()}",
                    )
                self.context_vars.append(source)

    def provide(
            self,
            source: Callable[..., Any] | type,
            *,
            scope: BaseScope | None = None,
            provides: Any = None,
            cache: bool = True,
    ) -> CompositeDependencySource:
        if scope is None:
            scope = self.scope
        composite = provide_on_instance(
            source=source,
            scope=scope,
            provides=provides,
            cache=cache,
        )
        self._add_dependency_sources(str(source), composite.dependency_sources)
        return composite

    def provide_all(
            self,
            *provides: Any,
            scope: BaseScope | None = None,
            cache: bool = True,
    ) -> CompositeDependencySource:
        if scope is None:
            scope = self.scope
        composite = provide_all_on_instance(
            *provides, scope=scope, cache=cache,
        )
        self._add_dependency_sources("?", composite.dependency_sources)
        return composite

    def alias(
            self,
            *,
            source: type,
            provides: Any,
            cache: bool = True,
    ) -> CompositeDependencySource:
        composite = alias(
            source=source,
            provides=provides,
            cache=cache,
        )
        self._add_dependency_sources(str(source), composite.dependency_sources)
        return composite

    def decorate(
            self,
            source: Callable[..., Any] | type,
            *,
            provides: Any = None,
    ) -> CompositeDependencySource:
        composite = decorate_on_instance(
            source=source,
            provides=provides,
        )
        self._add_dependency_sources(str(source), composite.dependency_sources)
        return composite

    def to_component(self, component: Component) -> ProviderWrapper:
        return ProviderWrapper(component, self)

    def from_context(
            self, *, provides: Any, scope: BaseScope,
    ) -> CompositeDependencySource:
        composite = from_context(provides, scope=scope)
        self._add_dependency_sources(
            name=str(provides),
            sources=composite.dependency_sources,
        )
        return composite


class ProviderWrapper(BaseProvider):
    def __init__(self, component: Component, provider: Provider) -> None:
        super().__init__(component)
        self.factories.extend(provider.factories)
        self.aliases.extend(provider.aliases)
        self.decorators.extend(provider.decorators)



__all__ = (
    "FromDishka",
    "inject",
    "FastStreamProvider",
    "setup_dishka",
)

import warnings
from collections.abc import AsyncIterator, Awaitable, Callable
from contextlib import asynccontextmanager
from typing import Any, ParamSpec, TypeVar, cast

from faststream import BaseMiddleware, FastStream, context
from faststream.__about__ import __version__
from faststream.broker.message import StreamMessage
from faststream.types import DecodedMessage
from faststream.utils.context import ContextRepo

from dishka import AsyncContainer, FromDishka, Provider, Scope, from_context
from dishka.integrations.base import wrap_injection

T = TypeVar("T")
P = ParamSpec("P")

class FastStreamProvider(Provider):
    context = from_context(ContextRepo, scope=Scope.REQUEST)
    message = from_context(StreamMessage, scope=Scope.REQUEST)


FASTSTREAM_OLD_MIDDLEWARES = __version__ < "0.5"


class _DishkaBaseMiddleware(BaseMiddleware):
    def __init__(self, container: AsyncContainer) -> None:
        self.container = container

    def __call__(self, msg: Any | None = None) -> "_DishkaBaseMiddleware":
        self.msg = msg
        return self


if FASTSTREAM_OLD_MIDDLEWARES:

    class DishkaMiddleware(_DishkaBaseMiddleware):
        @asynccontextmanager
        async def consume_scope(  # type: ignore[override]
                self,
                *args: Any,
                **kwargs: Any,
        ) -> AsyncIterator[DecodedMessage]:
            async with self.container() as request_container:
                with context.scope("dishka", request_container):
                    async with super().consume_scope(  # type: ignore[attr-defined]
                        *args,
                        **kwargs,
                    ) as result:
                        yield result

else:

    class DishkaMiddleware(_DishkaBaseMiddleware):  # type: ignore[no-redef]
        async def consume_scope(
                self,
                call_next: Callable[[Any], Awaitable[Any]],
                msg: StreamMessage[Any],
        ) -> AsyncIterator[DecodedMessage]:
            async with self.container(
                {
                    type(msg): msg,
                    StreamMessage: msg,
                    ContextRepo: context,
                },
            ) as request_container:
                with context.scope("dishka", request_container):
                    return cast(
                        AsyncIterator[DecodedMessage],
                        await call_next(msg),
                    )


def setup_dishka(
        container: AsyncContainer,
        app: FastStream,
        *,
        finalize_container: bool = True,
        auto_inject: bool = False,
) -> None:
    assert app.broker, "You can't patch FastStream application without broker"  # noqa: S101

    if finalize_container:
        app.after_shutdown(container.close)

    if FASTSTREAM_OLD_MIDDLEWARES:
        app.broker.middlewares = (  # type: ignore[attr-defined]
            DishkaMiddleware(container),
            *app.broker.middlewares,  # type: ignore[attr-defined]
        )

        if auto_inject:
            warnings.warn(
                """
Auto injection is not supported for FastStream version less than 0.5.0
Please, update your FastStream installation
or use @inject at each subscriber manually.
            """,
                category=RuntimeWarning,
                stacklevel=1,
            )

    else:
        app.broker._middlewares = (  # noqa: SLF001
            DishkaMiddleware(container),
            *app.broker._middlewares,  # noqa: SLF001
        )

        for subscriber in app.broker._subscribers.values():  # noqa: SLF001
            subscriber._broker_middlewares = (  # noqa: SLF001
                DishkaMiddleware(container),
                *subscriber._broker_middlewares,  # noqa: SLF001
            )

        for publisher in app.broker._publishers.values():  # noqa: SLF001
            publisher._broker_middlewares = (  # noqa: SLF001
                DishkaMiddleware(container),
                *publisher._broker_middlewares,  # noqa: SLF001
            )

        if auto_inject:
            app.broker._call_decorators = (  # noqa: SLF001
                inject,
                *app.broker._call_decorators,  # noqa: SLF001
            )


def inject(func: Callable[P, T]) -> Callable[P, T]:
    return wrap_injection(
        func=func,
        is_async=True,
        container_getter=lambda *_: context.get_local("dishka"),
    )



from abc import ABC, ABCMeta
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Final,
    Generic,
    Protocol,
    TypeAlias,
    TypeVar,
)

from dishka._adaptix.common import TypeHint
from dishka._adaptix.feature_requirement import HAS_PY_311
from dishka._adaptix.type_tools.basic_utils import is_parametrized
from dishka._adaptix.type_tools.fundamentals import (
    get_generic_args,
    get_type_vars,
    strip_alias,
)
from dishka.entities.provides_marker import ProvideMultiple

__all__ = ["WithParents", "ParentsResolver"]

IGNORE_TYPES: Final = (
    type,
    object,
    Enum,
    ABC,
    ABCMeta,
    Generic,
    Protocol,
    Exception,
    BaseException,
)
TypeVarsMap: TypeAlias = dict[TypeHint, TypeHint]

if HAS_PY_311:
    def is_type_var_tuple(obj: TypeHint) -> bool:
        return getattr(obj, "__typing_is_unpacked_typevartuple__", False)
else:
    def is_type_var_tuple(obj: TypeHint) -> bool:
        return False


def has_orig_bases(obj: TypeHint) -> bool:
    return hasattr(obj, "__orig_bases__")


def is_ignored_type(origin_type: TypeHint) -> bool:
    return origin_type in IGNORE_TYPES


def create_type_vars_map(obj: TypeHint) -> dict[TypeHint, TypeHint]:
    origin_obj = strip_alias(obj)
    type_vars = list(get_type_vars(origin_obj) or get_type_vars(obj))
    if not type_vars:
        return {}

    type_vars_map = {}
    arguments = list(get_generic_args(obj))
    reversed_arguments = False
    while True:
        if len(type_vars) == 0:
            break

        type_var = type_vars[0]
        if isinstance(type_var, TypeVar):
            del type_vars[0]
            type_vars_map[type_var] = arguments.pop(0)
        else:
            if len(type_vars) == 1:
                if reversed_arguments:
                    arguments.reverse()
                type_vars_map[type_var] = arguments
                break
            type_vars.reverse()
            arguments.reverse()
            reversed_arguments = not reversed_arguments

    return type_vars_map



class ParentsResolver:
    def get_parents(self, child_type: TypeHint) -> list[TypeHint]:
        if is_ignored_type(strip_alias(child_type)):
            raise ValueError(
                f"The starting class {child_type!r} is in ignored types",
            )
        if is_parametrized(child_type) or has_orig_bases(child_type):
            return self._get_parents_for_generic(child_type)
        return self._get_parents_for_mro(child_type)

    def _get_parents_for_generic(
        self, child_type: TypeHint,
    ) -> list[TypeHint]:
        parents: list[TypeHint] = []
        self._recursion_get_parents(
            child_type=child_type,
            parents=parents,
            type_vars_map={},
        )
        return parents

    def _recursion_get_parents(
        self,
        child_type: TypeHint,
        parents: list[TypeHint],
        type_vars_map: TypeVarsMap,
    ) -> None:
        origin_child_type = strip_alias(child_type)
        parametrized = is_parametrized(child_type)
        orig_bases = has_orig_bases(origin_child_type)
        if not orig_bases and not parametrized:
            parents.extend(
                self._get_parents_for_mro(origin_child_type),
            )
            return

        new_type_vars_map = create_type_vars_map(child_type)
        new_type_vars_map.update(type_vars_map)
        parents.append(
            self._create_type(
                obj=child_type,
                type_vars_map=new_type_vars_map,
            ),
        )
        if not orig_bases:
            return
        for parent_type in origin_child_type.__orig_bases__:
            origin_parent_type = strip_alias(parent_type)
            if is_ignored_type(origin_parent_type):
                continue

            self._recursion_get_parents(
                child_type=parent_type,
                parents=parents,
                type_vars_map=new_type_vars_map,
            )

    def _get_parents_for_mro(
        self, child_type: TypeHint,
    ) -> list[TypeHint]:
        return [
            parent_type for parent_type in child_type.mro()
            if not is_ignored_type(strip_alias(parent_type))
        ]

    def _create_type(
        self,
        obj: TypeHint,
        type_vars_map: TypeVarsMap,
    ) -> TypeHint:
        origin_obj = strip_alias(obj)
        type_vars = get_type_vars(origin_obj) or get_type_vars(obj)
        if not type_vars:
            return obj

        generic_args = []
        for type_var in type_vars:
            arg = type_vars_map[type_var]
            if isinstance(arg, list):
                generic_args.extend(arg)
            else:
                generic_args.append(arg)
        return origin_obj[tuple(generic_args)]


if TYPE_CHECKING:
    from typing import Union as WithParents
else:
    class WithParents:
        def __class_getitem__(cls, item: TypeHint) -> TypeHint:
            parents = ParentsResolver().get_parents(item)
            if len(parents) > 1:
                return ProvideMultiple(parents)
            return parents[0]



__all__ = [
    "AnyOf",
    "AsyncContainer",
    "BaseScope",
    "Component",
    "Container",
    "DEFAULT_COMPONENT",
    "DependencyKey",
    "FromComponent",
    "FromDishka",
    "Provider",
    "Scope",
    "WithParents",
    "alias",
    "decorate",
    "from_context",
    "make_async_container",
    "make_container",
    "provide",
    "provide_all",
    "new_scope",
]

from .async_container import AsyncContainer, make_async_container
from .container import Container, make_container
from .dependency_source import (
    alias,
    decorate,
    from_context,
    provide,
    provide_all,
)
from .entities.component import DEFAULT_COMPONENT, Component
from .entities.depends_marker import FromDishka
from .entities.key import DependencyKey, FromComponent
from .entities.provides_marker import AnyOf
from .entities.scope import BaseScope, Scope, new_scope
from .entities.with_parents import WithParents
from .provider import Provider



__all__ = [
    "FromDishka",
    "inject",
    "setup_dishka",
]

from collections.abc import Callable
from typing import Final, TypeVar

from click import (
    Command,
    Context,
    Group,
    get_current_context,
)

from dishka import Container, FromDishka
from .base import is_dishka_injected, wrap_injection

T = TypeVar("T")
CONTAINER_NAME: Final = "dishka_container"


def inject(func: Callable[..., T]) -> Callable[..., T]:
    return wrap_injection(
        func=func,
        container_getter=lambda _, __: get_current_context().meta[
            CONTAINER_NAME
        ],
        remove_depends=True,
        is_async=False,
    )


def _inject_commands(context: Context, command: Command | None) -> None:
    if isinstance(command, Command) and not is_dishka_injected(
        command.callback,  # type: ignore[arg-type]
    ):
        command.callback = inject(command.callback)   # type: ignore[arg-type]

    if isinstance(command, Group):
        for command_name in command.list_commands(context):
            child_command = command.get_command(context, command_name)
            _inject_commands(context, child_command)


def setup_dishka(
    container: Container,
    context: Context,
    *,
    finalize_container: bool = True,
    auto_inject: bool = False,
) -> None:
    context.meta[CONTAINER_NAME] = container

    if finalize_container:
        context.call_on_close(container.close)

    if auto_inject:
        _inject_commands(context, context.command)



from collections.abc import Sequence

from dishka.entities.key import DependencyKey
from .dependency_source import Factory

try:
    from builtins import ExceptionGroup  # type: ignore[attr-defined]

except ImportError:
    from exceptiongroup import (  # type: ignore[no-redef, import-not-found]
        ExceptionGroup,
    )

from .error_rendering import PathRenderer

_renderer = PathRenderer()


class DishkaError(Exception):
    pass


class InvalidGraphError(DishkaError):
    pass


class UnknownScopeError(InvalidGraphError):
    pass


class CycleDependenciesError(InvalidGraphError):
    def __init__(self, path: Sequence[Factory]) -> None:
        self.path = path

    def __str__(self) -> str:
        if len(self.path) == 1:
            hint = " Did you mean @decorate instead of @provide?"
        else:
            hint = ""
        details = _renderer.render(self.path)
        return f"Cycle dependencies detected.{hint}\n{details}"


class ExitError(ExceptionGroup[Exception], DishkaError):
    pass


class UnsupportedFactoryError(DishkaError):
    pass


class NoContextValueError(DishkaError):
    pass


class NoFactoryError(DishkaError):
    def __init__(
            self,
            requested: DependencyKey,
            path: Sequence[Factory] = (),
    ) -> None:
        self.requested = requested
        self.path = list(path)

    def add_path(self, requested_by: Factory) -> None:
        self.path.insert(0, requested_by)

    def __str__(self) -> str:
        if self.path:
            return (
                f"Cannot find factory for {self.requested}. "
                f"It is missing or has invalid scope.\n"
            ) + _renderer.render(self.path, self.requested)
        else:
            return (
                f"Cannot find factory for {self.requested}. "
                f"Check scopes in your providers. "
                f"It is missing or has invalid scope."
            )


class GraphMissingFactoryError(NoFactoryError, InvalidGraphError):
    pass



import warnings
from typing import TYPE_CHECKING, Annotated, TypeVar

from .component import DEFAULT_COMPONENT, Component
from .key import FromComponent

T = TypeVar("T")

if TYPE_CHECKING:
    from typing import Union
    FromDishka = Union[T, T]  # noqa: UP007,PYI016
else:
    class FromDishka:
        def __init__(self, component: Component = None):
            if component is None:
                self.component = DEFAULT_COMPONENT
                warnings.warn(
                    "Annotated[Cls, FromDishka()] is deprecated "
                    "use `FromDishka[Cls]` or "
                    "`Annotated[Cls, FromComponent()]` instead",
                    DeprecationWarning,
                    stacklevel=2,
                )

            else:
                self.component = component
                warnings.warn(
                    "Annotated[Cls, FromDishka(component)] is deprecated "
                    "use `Annotated[Cls, FromComponent(component)]` instead",
                    DeprecationWarning,
                    stacklevel=2,
                )

        def __class_getitem__(cls, item: T) -> Annotated[T, FromComponent]:
            return Annotated[item, FromComponent()]



import types
from typing import TypeVar, get_args, get_origin, get_type_hints

from ..common import TypeHint, VarTuple
from ..feature_requirement import HAS_ANNOTATED, HAS_SUPPORTED_PYDANTIC_PKG

__all__ = ("is_pydantic_class", "strip_alias", "get_type_vars", "get_generic_args", "get_all_type_hints")


if HAS_SUPPORTED_PYDANTIC_PKG:
    from pydantic import BaseModel

    _PYDANTIC_MCS = type(types.new_class("_PydanticSample", (BaseModel,), {}))

    def is_pydantic_class(tp) -> bool:
        return isinstance(tp, _PYDANTIC_MCS) and tp != BaseModel
else:
    def is_pydantic_class(tp) -> bool:
        return False


def strip_alias(tp: TypeHint) -> TypeHint:
    origin = tp.__pydantic_generic_metadata__["origin"] if is_pydantic_class(tp) else get_origin(tp)
    return tp if origin is None else origin


def get_type_vars(tp: TypeHint) -> VarTuple[TypeVar]:
    if is_pydantic_class(tp):
        return tp.__pydantic_generic_metadata__["parameters"]

    type_vars = getattr(tp, "__parameters__", ())
    # UnionType object contains descriptor inside `__parameters__`
    if not isinstance(type_vars, tuple):
        return ()
    return type_vars


def get_generic_args(tp: TypeHint) -> VarTuple[TypeHint]:
    if is_pydantic_class(tp):
        return tp.__pydantic_generic_metadata__["args"]
    return get_args(tp)


if HAS_ANNOTATED:
    def get_all_type_hints(obj, globalns=None, localns=None):
        return get_type_hints(obj, globalns, localns, include_extras=True)
else:
    get_all_type_hints = get_type_hints



__all__ = [
    "DependencySource",
    "alias", "Alias",
    "context_var", "ContextVariable", "from_context",
    "decorate", "Decorator",
    "provide", "provide_all",
    "Factory", "FactoryType",
]

from .alias import Alias
from .composite import DependencySource
from .context_var import ContextVariable
from .decorator import Decorator
from .factory import Factory, FactoryType
from .make_alias import alias
from .make_context_var import from_context
from .make_decorator import decorate
from .make_factory import provide, provide_all



__all__ = [
    "AutoInjectMiddleware",
    "CONTAINER_NAME",
    "FromDishka",
    "inject",
    "setup_dishka",
]

from collections.abc import Awaitable, Callable, Container
from inspect import Parameter
from typing import Any, Final, ParamSpec, TypeVar, cast

from aiogram import BaseMiddleware, Router
from aiogram.dispatcher.event.handler import HandlerObject
from aiogram.types import TelegramObject

from dishka import AsyncContainer, FromDishka
from .base import is_dishka_injected, wrap_injection

P = ParamSpec("P")
T = TypeVar("T")
CONTAINER_NAME: Final = "dishka_container"


def inject(func: Callable[P, T]) -> Callable[P, T]:
    additional_params = [Parameter(
        name=CONTAINER_NAME,
        annotation=Container,
        kind=Parameter.KEYWORD_ONLY,
    )]

    return wrap_injection(
        func=func,
        is_async=True,
        additional_params=additional_params,
        container_getter=lambda args, kwargs: kwargs[CONTAINER_NAME],
    )


class ContainerMiddleware(BaseMiddleware):
    def __init__(self, container: AsyncContainer) -> None:
        self.container = container

    async def __call__(
        self,
        handler: Callable[[TelegramObject, dict[str, Any]], Awaitable[Any]],
        event: TelegramObject,
        data: dict[str, Any],
    ) -> Any:
        async with self.container({TelegramObject: event}) as sub_container:
            data[CONTAINER_NAME] = sub_container
            return await handler(event, data)


class AutoInjectMiddleware(BaseMiddleware):
    async def __call__(
        self,
        handler: Callable[[TelegramObject, dict[str, Any]], Awaitable[Any]],
        event: TelegramObject,
        data: dict[str, Any],
    ) -> Any:
        old_handler = cast(HandlerObject, data["handler"])
        if is_dishka_injected(old_handler.callback):
            return await handler(event, data)

        new_handler = HandlerObject(
            callback=inject(old_handler.callback),
            filters=old_handler.filters,
            flags=old_handler.flags,
        )
        old_handler.callback = new_handler.callback
        old_handler.params = new_handler.params
        old_handler.varkw = new_handler.varkw
        old_handler.awaitable = new_handler.awaitable
        return await handler(event, data)


def setup_dishka(
    container: AsyncContainer,
    router: Router,
    *,
    auto_inject: bool = False,
) -> None:
    middleware = ContainerMiddleware(container)
    auto_inject_middleware = AutoInjectMiddleware()

    for observer in router.observers.values():
        observer.outer_middleware(middleware)
        if auto_inject and observer.event_name != "update":
            observer.middleware(auto_inject_middleware)



from typing import TYPE_CHECKING, Any, Callable, Tuple, Type, TypeVar, Union

K_contra = TypeVar("K_contra", contravariant=True)
V_co = TypeVar("V_co", covariant=True)
T = TypeVar("T")

Loader = Callable[[Any], V_co]
Dumper = Callable[[K_contra], Any]
Converter = Callable[..., Any]
Coercer = Callable[[Any, Any], Any]
OneArgCoercer = Callable[[Any], Any]

TypeHint = Any

VarTuple = Tuple[T, ...]

Catchable = Union[Type[BaseException], VarTuple[Type[BaseException]]]

# https://github.com/python/typing/issues/684#issuecomment-548203158
if TYPE_CHECKING:
    EllipsisType = ellipsis  # noqa: F821
else:
    EllipsisType = type(Ellipsis)



from abc import abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Protocol

from .dependency_source import FactoryType


@dataclass(slots=True)
class Exit:
    type: FactoryType
    callable: Callable[..., Any]


class CompiledFactory(Protocol):
    @abstractmethod
    def __call__(
            self,
            getter: Callable[..., Any],
            exits: list[Exit],
            context: Any,
    ) -> Any:
        raise NotImplementedError



from typing import Any

from dishka.entities.component import DEFAULT_COMPONENT
from dishka.entities.key import DependencyKey
from dishka.entities.scope import BaseScope
from .composite import CompositeDependencySource
from .context_var import ContextVariable, context_stub


def from_context(
        provides: Any, *, scope: BaseScope | None = None,
) -> CompositeDependencySource:
    composite = CompositeDependencySource(origin=context_stub)
    composite.dependency_sources.append(
        ContextVariable(
            provides=DependencyKey(
                type_hint=provides,
                component=DEFAULT_COMPONENT,
            ),
            scope=scope,
        ),
    )
    return composite



import typing
from dataclasses import dataclass, replace
from itertools import chain
from typing import Callable, Collection, Dict, Generic, Hashable, Mapping, TypeVar

from ..common import TypeHint
from ..feature_requirement import HAS_TV_TUPLE, HAS_UNPACK
from . import get_generic_args
from .basic_utils import get_type_vars, get_type_vars_of_parametrized, is_generic, is_parametrized, strip_alias
from .implicit_params import fill_implicit_params
from .normalize_type import normalize_type

M = TypeVar("M")
K = TypeVar("K", bound=Hashable)


@dataclass
class MembersStorage(Generic[K, M]):
    members: Mapping[K, TypeHint]
    overriden: Collection[K]
    meta: M


class GenericResolver(Generic[K, M]):
    def __init__(self, members_getter: Callable[[TypeHint], MembersStorage[K, M]]):
        self._raw_members_getter = members_getter

    def get_resolved_members(self, tp: TypeHint) -> MembersStorage[K, M]:
        if is_parametrized(tp):
            return self._get_members_of_parametrized_generic(tp)
        if is_generic(tp):
            return self._get_members_of_parametrized_generic(fill_implicit_params(tp))
        return self._get_members_by_parents(tp)

    def _get_members_of_parametrized_generic(self, parametrized_generic) -> MembersStorage[K, M]:
        origin = strip_alias(parametrized_generic)
        members_storage = self._get_members_by_parents(origin)
        type_var_to_actual = self._get_type_var_to_actual(
            get_type_vars(origin),
            self._unpack_args(get_generic_args(parametrized_generic)),
        )
        return replace(
            members_storage,
            members={
                key: self._parametrize_by_dict(type_var_to_actual, tp)
                for key, tp in members_storage.members.items()
            },
        )

    def _unpack_args(self, args):
        if HAS_UNPACK and any(strip_alias(arg) == typing.Unpack for arg in args):
            return tuple(arg.source for arg in normalize_type(tuple[args]).args)
        return args

    def _get_type_var_to_actual(self, type_vars, args):
        result = {}
        idx = 0
        for tv in type_vars:
            if HAS_TV_TUPLE and isinstance(tv, typing.TypeVarTuple):
                tuple_len = len(args) - len(type_vars) + 1
                result[tv] = args[idx:idx + tuple_len]
                idx += tuple_len
            else:
                result[tv] = (args[idx], )
                idx += 1

        return result

    def _parametrize_by_dict(self, type_var_to_actual, tp: TypeHint) -> TypeHint:
        if tp in type_var_to_actual:
            return type_var_to_actual[tp][0]

        params = get_type_vars_of_parametrized(tp)
        if not params:
            return tp
        return tp[tuple(chain.from_iterable(type_var_to_actual[type_var] for type_var in params))]

    def _get_members_by_parents(self, tp) -> MembersStorage[K, M]:
        members_storage = self._raw_members_getter(tp)
        if not any(
            get_type_vars_of_parametrized(tp) or isinstance(tp, TypeVar)
            for tp in members_storage.members.values()
        ):
            return members_storage
        if not hasattr(tp, "__orig_bases__"):
            return members_storage

        bases_members: Dict[K, TypeHint] = {}
        for base in reversed(tp.__orig_bases__):
            bases_members.update(self.get_resolved_members(base).members)

        return replace(
            members_storage,
            members={
                key: (
                    bases_members[key]
                    if (
                        key in bases_members
                        and key not in members_storage.overriden
                        and (is_generic(value) or isinstance(value, TypeVar))
                    )
                    else value
                )
                for key, value in members_storage.members.items()
            },
        )



# ruff: noqa: RET503
import dataclasses
import sys
import types
import typing
from abc import ABC, abstractmethod
from collections import abc as c_abc, defaultdict
from copy import copy
from dataclasses import InitVar, dataclass
from enum import Enum, EnumMeta
from functools import lru_cache, partial
from typing import (
    Any,
    Callable,
    ClassVar,
    DefaultDict,
    Dict,
    Final,
    ForwardRef,
    Hashable,
    Iterable,
    List,
    Literal,
    NewType,
    NoReturn,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from ..common import TypeHint, VarTuple
from ..feature_requirement import (
    HAS_ANNOTATED,
    HAS_PARAM_SPEC,
    HAS_PY_310,
    HAS_PY_311,
    HAS_SELF_TYPE,
    HAS_TV_SYNTAX,
    HAS_TV_TUPLE,
    HAS_TYPE_ALIAS,
    HAS_TYPE_GUARD,
    HAS_TYPE_UNION_OP,
    HAS_TYPED_DICT_REQUIRED,
    HAS_UNPACK,
)
from .basic_utils import create_union, eval_forward_ref, is_new_type, is_subclass_soft, strip_alias
from .fundamentals import get_generic_args
from .implicit_params import ImplicitParamsGetter


class BaseNormType(Hashable, ABC):
    @property
    @abstractmethod
    def origin(self) -> Any:
        ...

    @property
    @abstractmethod
    def args(self) -> VarTuple[Any]:
        ...

    @property
    @abstractmethod
    def source(self) -> TypeHint:
        ...


T = TypeVar("T")


class _BasicNormType(BaseNormType, ABC):
    __slots__ = ("_source", "_args")

    def __init__(self, args: VarTuple[Any], *, source: TypeHint):
        self._source = source
        self._args = args

    @property
    def args(self) -> VarTuple[Any]:
        return self._args

    @property
    def source(self) -> TypeHint:
        return self._source

    def __hash__(self):
        return hash((self.origin, self._args))

    def __eq__(self, other):
        if isinstance(other, _BasicNormType):
            return (
                self.origin == other.origin
                and
                self._args == other._args
            )
        if isinstance(other, BaseNormType):
            return False
        return NotImplemented

    def __repr__(self):
        args_str = f" {list(self.args)}," if self.args else ""
        return f"<{type(self).__name__}({self.origin},{args_str} source={self._source})>"


class _NormType(_BasicNormType):
    __slots__ = (*_BasicNormType.__slots__, "_source")

    def __init__(self, origin: TypeHint, args: VarTuple[Any], *, source: TypeHint):
        self._origin = origin
        super().__init__(args, source=source)

    @property
    def origin(self) -> Any:
        return self._origin


class _UnionNormType(_BasicNormType):
    def __init__(self, args: VarTuple[Any], *, source: TypeHint):
        super().__init__(self._order_args(args), source=source)

    @property
    def origin(self) -> Any:
        return Union

    # ensure stable order of args during one interpreter session
    def _make_orderable(self, obj: object) -> str:
        if isinstance(obj, BaseNormType):
            return f"{obj.origin} {[self._make_orderable(arg) for arg in obj.args]}"
        return str(obj)

    def _order_args(self, args: VarTuple[BaseNormType]) -> VarTuple[BaseNormType]:
        args_list = list(args)
        args_list.sort(key=self._make_orderable)
        return tuple(args_list)


def _type_and_value_iter(args):
    return [(type(arg), arg) for arg in args]


LiteralArg = Union[str, int, bytes, Enum]


class _LiteralNormType(_BasicNormType):
    def __init__(self, args: VarTuple[Any], *, source: TypeHint):
        super().__init__(self._order_args(args), source=source)

    @property
    def origin(self) -> Any:
        return Literal

    # ensure stable order of args during one interpreter session
    def _make_orderable(self, obj: LiteralArg) -> str:
        return f"{type(obj)}{obj.name}" if isinstance(obj, Enum) else repr(obj)

    def _order_args(self, args: VarTuple[LiteralArg]) -> VarTuple[LiteralArg]:
        args_list = list(args)
        args_list.sort(key=self._make_orderable)
        return tuple(args_list)

    def __eq__(self, other):
        if isinstance(other, _LiteralNormType):
            return _type_and_value_iter(self._args) == _type_and_value_iter(other._args)
        if isinstance(other, BaseNormType):
            return False
        return NotImplemented

    __hash__ = _BasicNormType.__hash__


class _AnnotatedNormType(_BasicNormType):
    @property
    def origin(self) -> Any:
        return typing.Annotated

    __slots__ = (*_BasicNormType.__slots__, "_hash")

    def __init__(self, args: VarTuple[Hashable], *, source: TypeHint):
        super().__init__(args, source=source)
        self._hash = self._calc_hash()

    # calculate hash even if one of Annotated metadata is not hashable
    def _calc_hash(self) -> int:
        lst = [self.origin]
        for arg in self._args:
            try:
                arg_hash = hash(arg)
            except TypeError:
                pass
            else:
                lst.append(arg_hash)
        return hash(tuple(lst))

    def __hash__(self):
        return self._hash


class Variance(Enum):
    INVARIANT = 0
    COVARIANT = 1
    CONTRAVARIANT = 2
    INFERRED = 3

    def __repr__(self):
        return f"{type(self).__name__}.{self.name}"


@dataclass
class Bound:
    value: BaseNormType


@dataclass
class Constraints:
    value: VarTuple[BaseNormType]


TypeVarLimit = Union[Bound, Constraints]


class NormTV(BaseNormType):
    __slots__ = ("_var", "_limit", "_variance", "_source")

    def __init__(self, var: Any, limit: TypeVarLimit, *, source: TypeHint):
        self._var = var
        self._source = source
        self._limit = limit

        if var.__covariant__:
            self._variance = Variance.COVARIANT
        if var.__contravariant__:
            self._variance = Variance.CONTRAVARIANT
        if getattr(var, "__infer_variance__", False):
            self._variance = Variance.INFERRED
        self._variance = Variance.INVARIANT

    @property
    def origin(self) -> Any:
        return self._var

    @property
    def args(self) -> Tuple[()]:
        return ()

    @property
    def source(self) -> TypeHint:
        return self._source

    @property
    def name(self) -> str:
        return self._var.__name__

    @property
    def variance(self) -> Variance:
        return self._variance

    @property
    def limit(self) -> TypeVarLimit:
        return self._limit

    def __repr__(self):
        return f"<{type(self).__name__}({self._var})>"

    def __hash__(self):
        return hash(self._var)

    def __eq__(self, other):
        if isinstance(other, NormTV):
            return self._var == other._var
        if isinstance(other, BaseNormType):
            return False
        return NotImplemented


class NormTVTuple(BaseNormType):
    __slots__ = ("_var", "_source")

    def __init__(self, var: Any, *, source: TypeHint):
        self._var = var
        self._source = source

    @property
    def origin(self) -> Any:
        return self._var

    @property
    def args(self) -> Tuple[()]:
        return ()

    @property
    def source(self) -> TypeHint:
        return self._source

    @property
    def name(self) -> str:
        return self._var.__name__

    def __repr__(self):
        return f"<{type(self).__name__}({self._var})>"

    def __hash__(self):
        return hash(self._var)

    def __eq__(self, other):
        if isinstance(other, NormTVTuple):
            return self._var == other._var
        if isinstance(other, BaseNormType):
            return False
        return NotImplemented


class NormParamSpecMarker(BaseNormType, ABC):
    __slots__ = ("_param_spec", "_source")

    def __init__(self, param_spec: Any, *, source: TypeHint):
        self._param_spec = param_spec
        self._source = source

    @property
    def param_spec(self) -> NormTV:
        return self._param_spec

    @property
    def args(self) -> Tuple[()]:
        return ()

    @property
    def source(self) -> TypeHint:
        return self._source

    def __hash__(self):
        return hash((self.origin, self.param_spec))

    def __eq__(self, other):
        if isinstance(other, NormParamSpecMarker):
            return self.origin == other.origin and self._param_spec == other._param_spec
        if isinstance(other, BaseNormType):
            return False
        return NotImplemented


class _NormParamSpecArgs(NormParamSpecMarker):
    @property
    def origin(self) -> Any:
        return typing.ParamSpecArgs


class _NormParamSpecKwargs(NormParamSpecMarker):
    @property
    def origin(self) -> Any:
        return typing.ParamSpecKwargs


AnyNormTypeVarLike = Union[NormTV, NormTVTuple]


class NormTypeAlias(BaseNormType):
    __slots__ = ("_type_alias", "_args", "_norm_type_vars")

    def __init__(self, type_alias, args: VarTuple[BaseNormType], type_vars: VarTuple[AnyNormTypeVarLike]):
        self._type_alias = type_alias
        self._args = args
        self._type_vars = type_vars

    @property
    def origin(self) -> Any:
        return self._type_alias

    @property
    def args(self) -> VarTuple[BaseNormType]:
        return self._args

    @property
    def source(self) -> TypeHint:
        return self._type_alias

    @property
    def value(self):
        return self._type_alias.__value__

    @property
    def module(self):
        return self._type_alias.__module__

    @property
    def type_params(self) -> VarTuple[AnyNormTypeVarLike]:
        return self._type_vars

    def __eq__(self, other):
        if isinstance(other, NormTypeAlias):
            return self._type_alias == other._type_alias and self._args == other._args
        if isinstance(other, BaseNormType):
            return False
        return NotImplemented

    def __hash__(self):
        return hash(self._type_alias)


_PARAM_SPEC_MARKER_TYPES = (typing.ParamSpecArgs, typing.ParamSpecKwargs) if HAS_PARAM_SPEC else ()


def make_norm_type(
    origin: TypeHint,
    args: VarTuple[Hashable],
    *,
    source: TypeHint,
) -> BaseNormType:
    if origin == Union:
        if not all(isinstance(arg, BaseNormType) for arg in args):
            raise TypeError
        return _UnionNormType(args, source=source)
    if origin == Literal:
        if not all(type(arg) in [int, bool, str, bytes] or isinstance(type(arg), EnumMeta) for arg in args):
            raise TypeError
        return _LiteralNormType(args, source=source)
    if HAS_ANNOTATED and origin == typing.Annotated:
        return _AnnotatedNormType(args, source=source)
    if isinstance(origin, TypeVar):
        raise TypeError
    if HAS_PARAM_SPEC and (
        isinstance(origin, _PARAM_SPEC_MARKER_TYPES) or isinstance(source, _PARAM_SPEC_MARKER_TYPES)
    ):
        raise TypeError
    return _NormType(origin, args, source=source)


NoneType = type(None)
ANY_NT = _NormType(Any, (), source=Any)


def _create_norm_union(args: VarTuple[BaseNormType]) -> BaseNormType:
    return _UnionNormType(args, source=create_union(tuple(a.source for a in args)))


def _dedup(inp: Iterable[T]) -> Iterable[T]:
    in_set = set()
    result = []
    for item in inp:
        if item not in in_set:
            result.append(item)
            in_set.add(item)
    return result


def _create_norm_literal(args: Iterable):
    dedup_args = tuple(_dedup(args))
    return _LiteralNormType(
        dedup_args,
        source=Literal[dedup_args],
    )


def _replace_source(norm: BaseNormType, *, source: TypeHint) -> BaseNormType:
    norm_copy = copy(norm)
    norm_copy._source = source  # type: ignore[attr-defined]
    return norm_copy


def _replace_source_with_union(norm: BaseNormType, sources: list) -> BaseNormType:
    return _replace_source(
        norm=norm,
        source=create_union(tuple(sources)),
    )


NormAspect = Callable[["TypeNormalizer", Any, Any, tuple], Optional[BaseNormType]]


class AspectStorage(List[str]):
    @overload
    def add(self, *, condition: object = True) -> Callable[[NormAspect], NormAspect]:
        ...

    @overload
    def add(self, func: NormAspect) -> NormAspect:
        ...

    def add(self, func: Optional[NormAspect] = None, *, condition: object = True) -> Any:
        if func is None:
            return partial(self.add, condition=condition)

        if condition:
            self.append(func.__name__)
        return func

    def copy(self) -> "AspectStorage":
        return type(self)(super().copy())


class NotSubscribedError(ValueError):
    pass


N = TypeVar("N", bound=BaseNormType)
TN = TypeVar("TN", bound="TypeNormalizer")


class TypeNormalizer:
    def __init__(self, implicit_params_getter: ImplicitParamsGetter):
        self.implicit_params_getter = implicit_params_getter
        self._namespace: Optional[Dict[str, Any]] = None

    def _with_namespace(self: TN, namespace: Dict[str, Any]) -> TN:
        self_copy = copy(self)
        self_copy._namespace = namespace
        return self_copy

    def _with_module_namespace(self: TN, module_name: str) -> TN:
        try:
            module = sys.modules[module_name]
        except KeyError:
            return self
        return self._with_namespace(vars(module))

    @overload
    def normalize(self, tp: TypeVar) -> NormTV:
        ...

    @overload
    def normalize(self, tp: TypeHint) -> BaseNormType:
        ...

    def normalize(self, tp: TypeHint) -> BaseNormType:
        origin = strip_alias(tp)
        args = get_generic_args(tp)

        result = self._norm_forward_ref(tp)
        if result is not None:
            return result

        for attr_name in self._aspect_storage:
            result = getattr(self, attr_name)(tp, origin, args)
            if result is not None:
                return result

        raise RuntimeError

    def _norm_forward_ref(self, tp):
        if self._namespace is not None:
            if isinstance(tp, str):
                return _replace_source(
                    self.normalize(ForwardRef(tp)),
                    source=tp,
                )
            if isinstance(tp, ForwardRef):
                return _replace_source(
                    self.normalize(eval_forward_ref(self._namespace, tp)),
                    source=tp,
                )
        elif isinstance(tp, (str, ForwardRef)):
            raise ValueError(f"Can not normalize value {tp!r}, there are no namespace to evaluate types")

    _aspect_storage = AspectStorage()

    def _norm_iter(self, tps: Iterable[Any]) -> VarTuple[BaseNormType]:
        return tuple(self.normalize(tp) for tp in tps)

    MUST_SUBSCRIBED_ORIGINS = [
        ClassVar, Final, Literal,
        Union, Optional, InitVar,
    ]
    if HAS_ANNOTATED:
        MUST_SUBSCRIBED_ORIGINS.append(typing.Annotated)
    if HAS_TYPE_GUARD:
        MUST_SUBSCRIBED_ORIGINS.append(typing.TypeGuard)
    if HAS_TYPED_DICT_REQUIRED:
        MUST_SUBSCRIBED_ORIGINS.extend([typing.Required, typing.NotRequired])

    @_aspect_storage.add
    def _check_bad_input(self, tp, origin, args):
        if tp in self.MUST_SUBSCRIBED_ORIGINS:
            raise NotSubscribedError(f"{tp} must be subscribed")

        if tp in (NewType, TypeVar):
            raise ValueError(f"{origin} must be instantiated")

    @_aspect_storage.add
    def _norm_none(self, tp, origin, args):
        if origin is None or origin is NoneType:
            return _NormType(None, (), source=tp)

    @_aspect_storage.add(condition=HAS_ANNOTATED)
    def _norm_annotated(self, tp, origin, args):
        if origin == typing.Annotated:
            return _AnnotatedNormType(
                (self.normalize(args[0]), *args[1:]),
                source=tp,
            )

    def _get_bound(self, type_var) -> Bound:
        return (
            Bound(ANY_NT)
            if (type_var.__bound__ is None or type_var.__bound__ is NoneType) else
            Bound(self.normalize(type_var.__bound__))
        )

    @_aspect_storage.add
    def _norm_type_var(self, tp, origin, args):
        if isinstance(origin, TypeVar):
            namespaced = self._with_module_namespace(origin.__module__)
            limit = (
                Constraints(
                    tuple(
                        namespaced._dedup_union_args(
                            namespaced._norm_iter(origin.__constraints__),
                        ),
                    ),
                )
                if origin.__constraints__ else
                namespaced._get_bound(origin)
            )
            return NormTV(var=origin, limit=limit, source=tp)

    @_aspect_storage.add(condition=HAS_TV_TUPLE)
    def _norm_type_var_tuple(self, tp, origin, args):
        if isinstance(origin, typing.TypeVarTuple):
            return NormTVTuple(var=origin, source=tp)

    @_aspect_storage.add(condition=HAS_PARAM_SPEC)
    def _norm_param_spec(self, tp, origin, args):
        if isinstance(tp, typing.ParamSpecArgs):
            return _NormParamSpecArgs(param_spec=self.normalize(origin), source=tp)

        if isinstance(tp, typing.ParamSpecKwargs):
            return _NormParamSpecKwargs(param_spec=self.normalize(origin), source=tp)

        if isinstance(origin, typing.ParamSpec):
            namespaced = self._with_module_namespace(origin.__module__)
            return NormTV(
                var=origin,
                limit=namespaced._get_bound(origin),
                source=tp,
            )

    @_aspect_storage.add(condition=HAS_TV_SYNTAX)
    def _norm_type_alias_type(self, tp, origin, args):
        if isinstance(origin, typing.TypeAliasType):
            return NormTypeAlias(
                origin,
                self._norm_iter(args),
                self._norm_iter(tp.__type_params__),
            )

    @_aspect_storage.add
    def _norm_init_var(self, tp, origin, args):
        if isinstance(origin, InitVar):
            # this origin is InitVar[T]
            return _NormType(
                InitVar,
                (self.normalize(origin.type),),
                source=tp,
            )

    @_aspect_storage.add
    def _norm_new_type(self, tp, origin, args):
        if is_new_type(tp):
            return _NormType(tp, (), source=tp)

    @_aspect_storage.add
    def _norm_tuple(self, tp, origin, args):
        if origin == tuple:
            if tp in (tuple, Tuple):  # not subscribed values
                return _NormType(
                    tuple,
                    (ANY_NT, ...),
                    source=tp,
                )

            # >>> Tuple[()].__args__ == ((),)
            # >>> tuple[()].__args__ == ()
            if not args or args == ((),):
                return _NormType(tuple, (), source=tp)

            if args[-1] is Ellipsis:
                return _NormType(
                    tuple, (*self._norm_iter(args[:-1]), ...),
                    source=tp,
                )

            norm_args = self._norm_iter(args)
            if HAS_UNPACK:
                norm_args = self._unpack_tuple_elements(norm_args)
            return _NormType(tuple, norm_args, source=tp)

    def _unpack_tuple_elements(self, args: VarTuple[BaseNormType]) -> VarTuple[BaseNormType]:
        # it is necessary to unpack the variable-length tuple as well
        if len(args) == 1 and args[0].origin == typing.Unpack:
            inner_tp = args[0].args[0]
            if inner_tp.origin == tuple:
                return inner_tp.args

        return self._unpack_generic_elements(args)

    def _unpack_generic_elements(self, args: VarTuple[Any]) -> VarTuple[BaseNormType]:
        result = []
        for arg in args:
            if isinstance(arg, BaseNormType) and arg.origin == typing.Unpack and self._is_fixed_size_tuple(arg.args[0]):
                result.extend(arg.args[0].args)
            else:
                result.append(arg)
        return tuple(result)

    def _is_fixed_size_tuple(self, tp: BaseNormType) -> bool:
        return tp.origin == tuple and (not tp.args or tp.args[-1] is not Ellipsis)

    @_aspect_storage.add
    def _norm_callable(self, tp, origin, args):
        if origin == c_abc.Callable:
            if not args:
                return _NormType(
                    c_abc.Callable, (..., ANY_NT), source=tp,
                )

            if args[0] is Ellipsis:
                call_args = ...
            elif isinstance(args[0], list):
                call_args = self._norm_iter(args[0])
                if HAS_TV_TUPLE:
                    call_args = self._unpack_generic_elements(call_args)
            else:
                call_args = normalize_type(args[0])
            return _NormType(
                c_abc.Callable, (call_args, self.normalize(args[-1])), source=tp,
            )

    @_aspect_storage.add
    def _norm_literal(self, tp, origin, args):
        if origin == Literal:
            if args == (None,):  # Literal[None] converted to None
                return _NormType(None, (), source=tp)

            if None in args:
                args_without_none = list(args)
                args_without_none.remove(None)

                return _UnionNormType(
                    (
                        _NormType(None, (), source=Literal[None]),
                        _create_norm_literal(args_without_none),
                    ),
                    source=tp,
                )

            return _LiteralNormType(args, source=tp)

    def _unfold_union_args(self, norm_args: Iterable[N]) -> Iterable[N]:
        result: List[N] = []
        for norm in norm_args:
            if norm.origin == Union:
                result.extend(norm.args)
            else:
                result.append(norm)
        return result

    def _dedup_union_args(self, args: Iterable[BaseNormType]) -> Iterable[BaseNormType]:
        args_to_sources: DefaultDict[BaseNormType, List[Any]] = defaultdict(list)

        for arg in args:
            args_to_sources[arg].append(arg.source)

        return [
            _replace_source_with_union(arg, sources)
            if len(sources) != 1 and isinstance(arg, _NormType)
            else arg
            for arg, sources in args_to_sources.items()
        ]

    def _merge_literals(self, args: Iterable[N]) -> Sequence[N]:
        result = []
        lit_args: List[N] = []
        for norm in args:
            if norm.origin == Literal:
                lit_args.extend(norm.args)
            else:
                result.append(norm)

        if lit_args:
            result.append(_create_norm_literal(lit_args))
        return result

    _UNION_ORIGINS: List[Any] = [Union]
    if HAS_TYPE_UNION_OP:
        _UNION_ORIGINS.append(types.UnionType)

    @_aspect_storage.add
    def _norm_union(self, tp, origin, args):
        if origin in self._UNION_ORIGINS:
            norm_args = self._norm_iter(args)
            unfolded_n_args = self._unfold_union_args(norm_args)
            unique_n_args = self._dedup_union_args(unfolded_n_args)
            merged_n_args = self._merge_literals(unique_n_args)

            if len(merged_n_args) == 1:
                arg = merged_n_args[0]
                return make_norm_type(origin=arg.origin, args=arg.args, source=tp)
            return _UnionNormType(tuple(merged_n_args), source=tp)

    @_aspect_storage.add
    def _norm_type(self, tp, origin, args):
        if is_subclass_soft(origin, type) and args:
            norm = self.normalize(args[0])

            if norm.origin == Union:
                return _UnionNormType(
                    tuple(
                        _NormType(type, (arg,), source=Type[arg.source])
                        for arg in norm.args
                    ),
                    source=tp,
                )

    ALLOWED_ZERO_PARAMS_ORIGINS = {Any, NoReturn}
    if HAS_TYPE_ALIAS:
        ALLOWED_ZERO_PARAMS_ORIGINS.add(typing.TypeAlias)
    if HAS_PY_310:
        ALLOWED_ZERO_PARAMS_ORIGINS.add(dataclasses.KW_ONLY)
    if HAS_PY_311:
        ALLOWED_ZERO_PARAMS_ORIGINS.add(typing.Never)
    if HAS_SELF_TYPE:
        ALLOWED_ZERO_PARAMS_ORIGINS.add(typing.Self)
    if HAS_PY_311:
        ALLOWED_ZERO_PARAMS_ORIGINS.add(typing.LiteralString)

    def _norm_generic_arg(self, arg):
        if arg is Ellipsis:
            return Ellipsis
        if isinstance(arg, tuple):
            return self._norm_iter(arg)
        return self.normalize(arg)

    def _norm_implicit_param(self, param):
        if param is Ellipsis:
            return Ellipsis
        return self.normalize(param)

    @_aspect_storage.add
    def _norm_other(self, tp, origin, args):
        if args:
            norm_args = tuple(self._norm_generic_arg(el) for el in args)
            if HAS_TV_TUPLE:
                norm_args = self._unpack_generic_elements(norm_args)
            return _NormType(origin, norm_args, source=tp)

        params = self.implicit_params_getter.get_implicit_params(origin)
        if not (
            params
            or isinstance(origin, type)
            or origin in self.ALLOWED_ZERO_PARAMS_ORIGINS
        ):
            raise ValueError(f"Can not normalize value {tp!r}")

        return _NormType(
            origin,
            tuple(self._norm_implicit_param(param) for param in params),
            source=tp,
        )


_STD_NORMALIZER = TypeNormalizer(ImplicitParamsGetter())
_cached_normalize = lru_cache(maxsize=128)(_STD_NORMALIZER.normalize)


def normalize_type(tp: TypeHint) -> BaseNormType:
    try:
        hash(tp)
    except TypeError:
        return _STD_NORMALIZER.normalize(tp)

    return _cached_normalize(tp)



from __future__ import annotations

import warnings
from asyncio import Lock
from collections.abc import Callable, MutableMapping
from types import TracebackType
from typing import Any, cast

from dishka.entities.component import DEFAULT_COMPONENT, Component
from dishka.entities.key import DependencyKey
from dishka.entities.scope import BaseScope, Scope
from .container_objects import Exit
from .context_proxy import ContextProxy
from .dependency_source import Factory, FactoryType
from .exceptions import (
    ExitError,
    NoFactoryError,
)
from .provider import BaseProvider
from .registry import Registry, RegistryBuilder


class AsyncContainer:
    __slots__ = (
        "child_registries",
        "close_parent",
        "lock",
        "parent_container",
        "registry",
        "_cache",
        "_context",
        "_exits",
    )

    def __init__(
            self,
            registry: Registry,
            *child_registries: Registry,
            parent_container: AsyncContainer | None = None,
            context: dict[Any, Any] | None = None,
            lock_factory: Callable[[], Lock] | None = None,
            close_parent: bool = False,
    ):
        self.registry = registry
        self.child_registries = child_registries
        self._context = {DependencyKey(type(self), DEFAULT_COMPONENT): self}
        if context:
            for key, value in context.items():
                if not isinstance(key, DependencyKey):
                    key = DependencyKey(key, DEFAULT_COMPONENT)
                self._context[key] = value
        self._cache = {**self._context}
        self.parent_container = parent_container

        self.lock: Lock | None
        if lock_factory:
            self.lock = lock_factory()
        else:
            self.lock = None
        self._exits: list[Exit] = []
        self.close_parent = close_parent

    @property
    def context(self) -> MutableMapping[DependencyKey, Any]:
        warnings.warn(
            "`container.context` is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        return ContextProxy(cache=self._cache, context=self._context)

    def __call__(
            self,
            context: dict[Any, Any] | None = None,
            lock_factory: Callable[[], Lock] | None = None,
            scope: BaseScope | None = None,
    ) -> AsyncContextWrapper:
        """
        Prepare container for entering the inner scope.
        :param context: Data which will available in inner scope
        :param lock_factory: Callable to create lock instance or None
        :param scope: target scope or None to enter next non-skipped scope
        :return: async context manager for inner scope
        """
        if not self.child_registries:
            raise ValueError("No child scopes found")

        child = AsyncContainer(
            *self.child_registries,
            parent_container=self,
            context=context,
            lock_factory=lock_factory,
        )
        if scope is None:
            while child.registry.scope.skip:
                if not child.child_registries:
                    raise ValueError("No non-skipped scopes found.")
                child = AsyncContainer(
                    *child.child_registries,
                    parent_container=child,
                    context=context,
                    lock_factory=lock_factory,
                    close_parent=True,
                )
        else:
            while child.registry.scope is not scope:
                if not child.child_registries:
                    raise ValueError(f"Cannot find {scope} as a child of "
                                     f"current {self.registry.scope}")
                child = AsyncContainer(
                    *child.child_registries,
                    parent_container=child,
                    context=context,
                    lock_factory=lock_factory,
                    close_parent=True,
                )
        return AsyncContextWrapper(child)

    async def get(
            self,
            dependency_type: Any,
            component: Component | None = DEFAULT_COMPONENT,
    ) -> Any:
        lock = self.lock
        key = DependencyKey(dependency_type, component)
        if not lock:
            return await self._get_unlocked(key)
        async with lock:
            return await self._get_unlocked(key)

    async def _get_unlocked(self, key: DependencyKey) -> Any:
        if key in self._cache:
            return self._cache[key]
        compiled = self.registry.get_compiled_async(key)
        if not compiled:
            if not self.parent_container:
                raise NoFactoryError(key)
            return await self.parent_container.get(
                key.type_hint, key.component,
            )
        try:
            return await compiled(self._get_unlocked, self._exits, self._cache)
        except NoFactoryError as e:
            # cast is needed because registry.get_factory will always
            # return Factory. This happens because registry.get_compiled
            # uses the same method and returns None if the factory is not found
            # If None is returned, then go to the parent container
            e.add_path(cast(Factory, self.registry.get_factory(key)))
            raise

    async def close(self, exception: BaseException | None = None) -> None:
        errors = []
        for exit_generator in self._exits[::-1]:
            try:
                if exit_generator.type is FactoryType.ASYNC_GENERATOR:
                    await exit_generator.callable.asend(exception) # type: ignore[attr-defined]
                elif exit_generator.type is FactoryType.GENERATOR:
                    exit_generator.callable.send(exception)  # type: ignore[attr-defined]
            except StopIteration:  # noqa: PERF203
                pass
            except StopAsyncIteration:
                pass
            except Exception as err:  # noqa: BLE001
                errors.append(err)
        self._cache = {**self._context}
        if self.close_parent and self.parent_container:
            try:
                await self.parent_container.close(exception)
            except Exception as err:  # noqa: BLE001
                errors.append(err)
        if errors:
            raise ExitError("Cleanup context errors", errors)


class AsyncContextWrapper:
    def __init__(self, container: AsyncContainer):
        self.container = container

    async def __aenter__(self) -> AsyncContainer:
        return self.container

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_val: BaseException | None = None,
        exc_tb: TracebackType | None = None,
    ) -> None:
        await self.container.close(exception=exc_val)


def make_async_container(
        *providers: BaseProvider,
        scopes: type[BaseScope] = Scope,
        context: dict[Any, Any] | None = None,
        lock_factory: Callable[[], Lock] | None = Lock,
        skip_validation: bool = False,
        start_scope: BaseScope | None = None,
) -> AsyncContainer:
    registries = RegistryBuilder(
        scopes=scopes,
        container_type=AsyncContainer,
        providers=providers,
        skip_validation=skip_validation,
    ).build()
    container = AsyncContainer(
        *registries,
        context=context,
        lock_factory=lock_factory,
    )

    if start_scope is None:
        while container.registry.scope.skip:
            container = AsyncContainer(
                *container.child_registries,
                parent_container=container,
                context=context,
                lock_factory=lock_factory,
                close_parent=True,
            )
    else:
        while container.registry.scope is not start_scope:
            container = AsyncContainer(
                *container.child_registries,
                parent_container=container,
                context=context,
                lock_factory=lock_factory,
                close_parent=True,
            )
    return container



from __future__ import annotations

from typing import Any

from dishka.entities.component import Component
from dishka.entities.key import DependencyKey
from dishka.entities.scope import BaseScope
from .factory import Factory, FactoryType


def _identity(x: Any) -> Any:
    return x


class Alias:
    __slots__ = ("source", "provides", "cache", "component")

    def __init__(
            self, *,
            source: DependencyKey,
            provides: DependencyKey,
            cache: bool,
    ) -> None:
        self.source = source
        self.provides = provides
        self.cache = cache

    def as_factory(
            self, scope: BaseScope | None, component: Component | None,
    ) -> Factory:
        return Factory(
            scope=scope,
            source=_identity,
            provides=self.provides.with_component(component),
            is_to_bind=False,
            dependencies=[self.source.with_component(component)],
            kw_dependencies={},
            type_=FactoryType.ALIAS,
            cache=self.cache,
        )

    def __get__(self, instance: Any, owner: Any) -> Alias:
        return self



__all__ = [
    "DishkaRoute",
    "FromDishka",
    "inject",
    "setup_dishka",
]

from collections.abc import Callable
from inspect import Parameter
from typing import Any, ParamSpec, TypeVar, get_type_hints

from fastapi import FastAPI, Request, WebSocket
from fastapi.routing import APIRoute

from dishka import AsyncContainer, FromDishka
from .base import wrap_injection
from .starlette import ContainerMiddleware

T = TypeVar("T")
P = ParamSpec("P")

def inject(func: Callable[P, T]) -> Callable[P, T]:
    hints = get_type_hints(func)
    request_hint = next(
        (name for name, hint in hints.items() if hint is Request),
        None,
    )
    websocket_hint = next(
        (name for name, hint in hints.items() if hint is WebSocket),
        None,
    )
    if request_hint is None and websocket_hint is None:
        additional_params = [
            Parameter(
                name="___dishka_request",
                annotation=Request,
                kind=Parameter.KEYWORD_ONLY,
            ),
        ]
    else:
        additional_params = []
    param_name = request_hint or websocket_hint or "___dishka_request"
    return wrap_injection(
        func=func,
        is_async=True,
        additional_params=additional_params,
        container_getter=lambda _, p: p[param_name].state.dishka_container,
    )


class DishkaRoute(APIRoute):
    def __init__(
        self,
        path: str,
        endpoint: Callable[..., Any],
        **kwargs: Any,
    ) -> None:
        endpoint = inject(endpoint)
        super().__init__(path, endpoint, **kwargs)


def setup_dishka(container: AsyncContainer, app: FastAPI) -> None:
    app.add_middleware(ContainerMiddleware)
    app.state.dishka_container = container



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

__all__ = ["AnyOf", "ProvideMultiple"]

if TYPE_CHECKING:
    from typing import Union as AnyOf
else:
    class AnyOf:
        def __class_getitem__(cls, item: Any) -> Any:
            if isinstance(item, tuple):
                return ProvideMultiple(item)
            return item


class ProvideMultiple:
    def __init__(self, items: Sequence[Any]) -> None:
        self.items = items



from collections import defaultdict
from collections.abc import Callable, Sequence
from typing import Any, TypeVar, cast, get_args, get_origin

from ._adaptix.type_tools.basic_utils import is_generic
from ._adaptix.type_tools.fundamentals import get_type_vars
from .container_objects import CompiledFactory
from .dependency_source import (
    Alias,
    ContextVariable,
    Decorator,
    Factory,
    FactoryType,
)
from .entities.component import DEFAULT_COMPONENT, Component
from .entities.key import DependencyKey
from .entities.scope import BaseScope, InvalidScopes, Scope
from .exceptions import (
    CycleDependenciesError,
    GraphMissingFactoryError,
    InvalidGraphError,
    NoFactoryError,
    UnknownScopeError,
)
from .factory_compiler import compile_factory
from .provider import BaseProvider


class UndecoratedType:
    """Container for a type which is decorated in other place."""
    def __init__(self, original: type[Any], depth: int) -> None:
        self.original = original
        self.level = depth

    def __repr__(self) -> str:
        return f"UndecoratedType({self.original}, depth={self.level})"


class Registry:
    __slots__ = ("scope", "factories", "compiled", "compiled_async")

    def __init__(self, scope: BaseScope):
        self.scope = scope
        self.factories: dict[DependencyKey, Factory] = {}
        self.compiled: dict[DependencyKey, Callable[..., Any]] = {}
        self.compiled_async: dict[DependencyKey, Callable[..., Any]] = {}

    def add_factory(self, factory: Factory) -> None:
        if is_generic(factory.provides.type_hint):
            origin = get_origin(factory.provides.type_hint)
            if origin:
                origin_key = DependencyKey(origin, factory.provides.component)
                self.factories[origin_key] = factory
        self.factories[factory.provides] = factory

    def get_compiled(
            self, dependency: DependencyKey,
    ) -> CompiledFactory | None:
        try:
            return self.compiled[dependency]
        except KeyError:
            factory = self.get_factory(dependency)
            if not factory:
                return None
            compiled = compile_factory(factory=factory, is_async=False)
            self.compiled[dependency] = compiled
            return compiled

    def get_compiled_async(
            self, dependency: DependencyKey,
    ) -> CompiledFactory | None:
        try:
            return self.compiled[dependency]
        except KeyError:
            factory = self.get_factory(dependency)
            if not factory:
                return None
            compiled = compile_factory(factory=factory, is_async=True)
            self.compiled[dependency] = compiled
            return compiled

    def get_factory(self, dependency: DependencyKey) -> Factory | None:
        try:
            return self.factories[dependency]
        except KeyError:
            origin = get_origin(dependency.type_hint)
            if not origin:
                return None
            origin_key = DependencyKey(origin, dependency.component)
            factory = self.factories.get(origin_key)
            if not factory:
                return None

            factory = self._specialize_generic(factory, dependency)
            self.factories[dependency] = factory
            return factory

    def _specialize_generic(
            self, factory: Factory, dependency_key: DependencyKey,
    ) -> Factory:
        dependency = dependency_key.type_hint
        type_var_deps = (
            d.type_hint
            for d in factory.dependencies
            if isinstance(d.type_hint, TypeVar)
        )
        params_replacement = dict(zip(
            type_var_deps,
            get_args(dependency),
            strict=False,
        ))
        new_dependencies: list[DependencyKey] = []
        for source_dependency in factory.dependencies:
            hint = source_dependency.type_hint
            if isinstance(hint, TypeVar):
                hint = params_replacement[hint]
            elif get_origin(hint):
                hint = hint[tuple(
                    params_replacement[param]
                    for param in get_type_vars(hint)
                )]
            new_dependencies.append(DependencyKey(
                hint, source_dependency.component,
            ))
        new_kw_dependencies: dict[str, DependencyKey] = {}
        for name, source_dependency in factory.kw_dependencies.items():
            hint = source_dependency.type_hint
            if isinstance(hint, TypeVar):
                hint = params_replacement[hint]
            elif get_origin(hint):
                hint = hint[tuple(
                    params_replacement[param]
                    for param in get_type_vars(hint)
                )]
            new_kw_dependencies[name] = DependencyKey(
                hint, source_dependency.component,
            )
        return Factory(
            source=factory.source,
            provides=dependency_key,
            dependencies=new_dependencies,
            kw_dependencies=new_kw_dependencies,
            is_to_bind=factory.is_to_bind,
            type_=factory.type,
            scope=factory.scope,
            cache=factory.cache,
        )


class GraphValidator:
    def __init__(self, registries: Sequence[Registry]) -> None:
        self.registries = registries
        self.path: dict[DependencyKey, Factory] = {}
        self.valid_keys: dict[DependencyKey, bool] = {}

    def _validate_key(
            self, key: DependencyKey, registry_index: int,
    ) -> None:
        if key in self.valid_keys:
            return
        if key in self.path:
            keys = list(self.path)
            factories = list(self.path.values())[keys.index(key):]
            raise CycleDependenciesError(factories)
        for index in range(registry_index + 1):
            registry = self.registries[index]
            factory = registry.get_factory(key)
            if factory:
                self._validate_factory(factory, registry_index)
                return
        raise NoFactoryError(requested=key)

    def _validate_factory(
            self, factory: Factory, registry_index: int,
    ) -> None:
        self.path[factory.provides] = factory
        if (
            factory.provides in factory.kw_dependencies.values() or
            factory.provides in factory.dependencies
        ):
            raise CycleDependenciesError([factory])
        try:
            for dep in factory.dependencies:
                # ignore TypeVar parameters
                if not isinstance(dep.type_hint, TypeVar):
                    self._validate_key(dep, registry_index)
            for dep in factory.kw_dependencies.values():
                # ignore TypeVar parameters
                if not isinstance(dep.type_hint, TypeVar):
                    self._validate_key(dep, registry_index)

        except NoFactoryError as e:
            e.add_path(factory)
            raise
        finally:
            self.path.pop(factory.provides)
        self.valid_keys[factory.provides] = True

    def validate(self) -> None:
        for registry_index, registry in enumerate(self.registries):
            factories = tuple(registry.factories.values())
            for factory in factories:
                self.path = {}
                try:
                    self._validate_factory(factory, registry_index)
                except NoFactoryError as e:
                    raise GraphMissingFactoryError(
                        e.requested, e.path,
                    ) from None
                except CycleDependenciesError as e:
                    raise e from None


class RegistryBuilder:
    def __init__(
            self,
            *,
            scopes: type[BaseScope],
            providers: Sequence[BaseProvider],
            container_type: type,
            skip_validation: bool,
    ) -> None:
        self.scopes = scopes
        self.providers = providers
        self.registries: dict[BaseScope, Registry] = {}
        self.dependency_scopes: dict[DependencyKey, BaseScope] = {}
        self.components: set[Component] = {DEFAULT_COMPONENT}
        self.alias_sources: dict[DependencyKey, Any] = {}
        self.aliases: dict[DependencyKey, Alias] = {}
        self.container_type = container_type
        self.decorator_depth: dict[DependencyKey, int] = defaultdict(int)
        self.skip_validation = skip_validation

    def _collect_components(self) -> None:
        for provider in self.providers:
            self.components.add(provider.component)

    def _collect_provided_scopes(self) -> None:
        for provider in self.providers:
            for factory in provider.factories:
                if not isinstance(factory.scope, self.scopes):
                    raise UnknownScopeError(
                        f"Scope {factory.scope} is unknown, "
                        f"expected one of {self.scopes}",
                    )
                provides = factory.provides.with_component(provider.component)
                self.dependency_scopes[provides] = factory.scope
            for context_var in provider.context_vars:
                for component in self.components:
                    provides = context_var.provides.with_component(component)
                    # typing.cast is applied because the scope
                    # was checked above
                    self.dependency_scopes[provides] = cast(
                        BaseScope, context_var.scope,
                    )

    def _collect_aliases(self) -> None:
        for provider in self.providers:
            component = provider.component
            for alias in provider.aliases:
                provides = alias.provides.with_component(component)
                alias_source = alias.source.with_component(component)
                self.alias_sources[provides] = alias_source
                self.aliases[provides] = alias

    def _init_registries(self) -> None:
        for scope in self.scopes:
            registry = Registry(scope)
            context_var = ContextVariable(
                provides=DependencyKey(self.container_type, DEFAULT_COMPONENT),
                scope=scope,
            )
            for component in self.components:
                registry.add_factory(context_var.as_factory(component))
            self.registries[scope] = registry

    def _process_factory(
            self, provider: BaseProvider, factory: Factory,
    ) -> None:
        registry = self.registries[cast(Scope, factory.scope)]
        registry.add_factory(factory.with_component(provider.component))

    def _process_alias(
            self, provider: BaseProvider, alias: Alias,
    ) -> None:
        component = provider.component
        alias_source = alias.source.with_component(component)
        visited_keys: list[DependencyKey] = []
        while alias_source not in self.dependency_scopes:
            if alias_source not in self.alias_sources:
                e = NoFactoryError(alias_source)
                for key in visited_keys[::-1]:
                    e.add_path(self.aliases[key].as_factory(
                        InvalidScopes.UNKNOWN_SCOPE, key.component,
                    ))
                e.add_path(alias.as_factory(
                    InvalidScopes.UNKNOWN_SCOPE, component,
                ))
                raise e
            visited_keys.append(alias_source)
            alias_source = self.alias_sources[alias_source]
            if alias_source in visited_keys:
                raise CycleDependenciesError([
                    self.aliases[s].as_factory(
                        InvalidScopes.UNKNOWN_SCOPE, component,
                    )
                    for s in visited_keys
                ])

        scope = self.dependency_scopes[alias_source]
        registry = self.registries[scope]

        factory = alias.as_factory(scope, component)
        self.dependency_scopes[factory.provides] = scope
        registry.add_factory(factory)

    def _process_decorator(
            self, provider: BaseProvider, decorator: Decorator,
    ) -> None:
        provides = decorator.provides.with_component(provider.component)
        if provides not in self.dependency_scopes:
            raise GraphMissingFactoryError(
                requested=provides,
                path=[decorator.as_factory(
                    scope=InvalidScopes.UNKNOWN_SCOPE,
                    new_dependency=provides,
                    cache=False,
                    component=provider.component,
                )],
            )
        scope = self.dependency_scopes[provides]
        registry = self.registries[scope]
        undecorated_type = UndecoratedType(
            decorator.provides.type_hint,
            self.decorator_depth[provides],
        )
        self.decorator_depth[provides] += 1
        old_factory = registry.get_factory(provides)
        if old_factory is None:
            raise InvalidGraphError(
                "Cannot apply decorator because there is"
                f"no factory for {provides}",
            )
        if old_factory.type is FactoryType.CONTEXT:
            raise InvalidGraphError(
                f"Cannot apply decorator to context data {provides}",
            )
        old_factory.provides = DependencyKey(
            undecorated_type, old_factory.provides.component,
        )
        new_factory = decorator.as_factory(
            scope=scope,
            new_dependency=DependencyKey(undecorated_type, None),
            cache=old_factory.cache,
            component=provider.component,
        )
        registry.add_factory(old_factory)
        registry.add_factory(new_factory)

    def _process_context_var(
            self, provider: BaseProvider, context_var: ContextVariable,
    ) -> None:
        if context_var.scope is None:
            raise UnknownScopeError(
                f"Scope {context_var.scope} is unknown, "
                f"expected one of {self.scopes}. Define it"
                f"explicitly in Provider or from_context",
            )
        registry = self.registries[context_var.scope]
        for component in self.components:
            registry.add_factory(context_var.as_factory(component))

    def build(self) -> tuple[Registry, ...]:
        self._collect_components()
        self._collect_provided_scopes()
        self._collect_aliases()
        self._init_registries()

        for provider in self.providers:
            for factory in provider.factories:
                self._process_factory(provider, factory)
            for alias in provider.aliases:
                self._process_alias(provider, alias)
            for context_var in provider.context_vars:
                self._process_context_var(provider, context_var)
            for decorator in provider.decorators:
                self._process_decorator(provider, decorator)

        registries = list(self.registries.values())
        if not self.skip_validation:
            GraphValidator(registries).validate()
        return tuple(registries)



from __future__ import annotations

from collections.abc import Sequence
from typing import Any, ClassVar

from .alias import Alias
from .context_var import ContextVariable
from .decorator import Decorator
from .factory import Factory

DependencySource = Alias | Factory | Decorator | ContextVariable


class CompositeDependencySource:
    _instances: ClassVar[int] = 0

    def __init__(
            self,
            origin: Any,
            dependency_sources: Sequence[DependencySource] = (),
    ) -> None:
        self.dependency_sources = list(dependency_sources)
        self.origin = origin
        self.number = self._instances
        CompositeDependencySource._instances += 1

    def __get__(self, instance: Any, owner: Any) -> CompositeDependencySource:
        try:
            origin = self.origin.__get__(instance, owner)
        except AttributeError:  # not a valid descriptor
            origin = self.origin
        return CompositeDependencySource(
            origin=origin,
            dependency_sources=[
                s.__get__(instance, owner) for s in self.dependency_sources
            ],
        )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.origin(*args, **kwargs)

    def __add__(
            self, other: CompositeDependencySource,
    ) -> CompositeDependencySource:
        return CompositeDependencySource(
            origin=None,
            dependency_sources=(
                self.dependency_sources
                + other.dependency_sources
            ),
        )


def ensure_composite(origin: Any) -> CompositeDependencySource:
    if isinstance(origin, CompositeDependencySource):
        return origin
    else:
        return CompositeDependencySource(origin)



__all__ = [
    "DISHKA_CONTAINER_KEY",
    "FromDishka",
    "inject",
    "setup_dishka",
]

from collections.abc import Callable, Coroutine
from typing import Any, Final, TypeAlias, TypeVar, cast

from aiohttp import web
from aiohttp.abc import AbstractView
from aiohttp.typedefs import Handler
from aiohttp.web_app import Application
from aiohttp.web_request import Request
from aiohttp.web_response import StreamResponse

from dishka import AsyncContainer, FromDishka, Scope
from dishka.integrations.base import is_dishka_injected, wrap_injection

T = TypeVar("T")
DISHKA_CONTAINER_KEY: Final = web.AppKey("dishka_container", AsyncContainer)
AiohttpHandler: TypeAlias = type[AbstractView] | Callable[
    [Request], Coroutine[Any, Any, StreamResponse],
]


def inject(func: Callable[..., T]) -> AiohttpHandler:
    return cast(
        AiohttpHandler,
        wrap_injection(
            func=func,
            is_async=True,
            container_getter=lambda p, _: p[0][DISHKA_CONTAINER_KEY],
        ),
    )


@web.middleware
async def container_middleware(
    request: Request,
    handler: Handler,
) -> StreamResponse:
    container = request.app[DISHKA_CONTAINER_KEY]

    if (
        request.headers.get("Upgrade") == "websocket"
        and request.headers.get("Connection") == "Upgrade"
    ):
        scope = Scope.SESSION

    else:
        scope = Scope.REQUEST

    context = {Request: request}

    async with container(context=context, scope=scope) as request_container:
        request[
            DISHKA_CONTAINER_KEY] = request_container  # type: ignore[index]
        return await handler(request)


def _inject_routes(router: web.UrlDispatcher) -> None:
    for route in router.routes():
        _inject_route(route)

    for resource in router.resources():
        for route in resource._routes:  # type: ignore[attr-defined] # noqa: SLF001
            _inject_route(route)


def _inject_route(route: web.AbstractRoute) -> None:
    if not is_dishka_injected(route.handler):
        # typing.cast is used because AbstractRoute._handler
        # is Handler or Type[AbstractView]
        route._handler = cast( # noqa: SLF001
            AiohttpHandler,
            inject(route.handler),
        )


async def _on_shutdown(app: web.Application) -> None:
    await app[DISHKA_CONTAINER_KEY].close()


def setup_dishka(
    container: AsyncContainer,
    app: Application,
    *,
    auto_inject: bool = False,
) -> None:
    app[DISHKA_CONTAINER_KEY] = container
    app.middlewares.append(container_middleware)
    app.on_shutdown.append(_on_shutdown)

    if auto_inject:
        _inject_routes(app.router)



from __future__ import annotations

from typing import Annotated, Any, NamedTuple, get_args, get_origin

from .component import DEFAULT_COMPONENT, Component


class FromComponent(NamedTuple):
    component: Component = DEFAULT_COMPONENT


class DependencyKey(NamedTuple):
    type_hint: Any
    component: Component | None

    def with_component(self, component: Component | None) -> DependencyKey:
        if self.component is not None:
            return self
        return DependencyKey(
            type_hint=self.type_hint,
            component=component,
        )

    def __str__(self) -> str:
        return f"({self.type_hint}, component={self.component!r})"


def hint_to_dependency_key(hint: Any) -> DependencyKey:
    if get_origin(hint) is not Annotated:
        return DependencyKey(hint, None)
    args = get_args(hint)
    from_component = next(
        (arg for arg in args if isinstance(arg, FromComponent)),
        None,
    )
    if from_component is None:
        return DependencyKey(args[0], None)
    return DependencyKey(args[0], from_component.component)


def hints_to_dependency_keys(hints: list[Any]) -> list[DependencyKey]:
    return [
        hint_to_dependency_key(type_hint)
        for type_hint in hints
    ]



__all__ = [
    "FromDishka",
    "inject",
    "setup_dishka",
]

from collections.abc import Callable
from typing import Any, ParamSpec, TypeVar, cast

from flask import Flask, Request, g, request
from flask.sansio.scaffold import Scaffold
from flask.typing import RouteCallable

from dishka import Container, FromDishka
from .base import is_dishka_injected, wrap_injection

T = TypeVar("T")
P = ParamSpec("P")


def inject(func: Callable[P, T]) -> Callable[P, T]:
    return wrap_injection(
        func=func,
        is_async=False,
        container_getter=lambda _, p: g.dishka_container,
    )


class ContainerMiddleware:
    def __init__(self, container: Container) -> None:
        self.container = container

    def enter_request(self) -> None:
        g.dishka_container_wrapper = self.container({Request: request})
        g.dishka_container = g.dishka_container_wrapper.__enter__()

    def exit_request(self, *_args: Any, **_kwargs: Any) -> None:
        g.dishka_container.close()


def _inject_routes(scaffold: Scaffold) -> None:
    for key, func in scaffold.view_functions.items():
        if not is_dishka_injected(func):
            # typing.cast is applied because there
            # are RouteCallable objects in dict value
            scaffold.view_functions[key] = cast(RouteCallable, inject(func))


def setup_dishka(
        container: Container,
        app: Flask,
        *,
        auto_inject: bool = False,
) -> None:
    middleware = ContainerMiddleware(container)
    app.before_request(middleware.enter_request)
    app.teardown_appcontext(middleware.exit_request)
    if auto_inject:
        _inject_routes(app)
        for blueprint in app.blueprints.values():
            _inject_routes(blueprint)



import types
import typing
from typing import Any, Dict, ForwardRef, Generic, NewType, Protocol, TypedDict, TypeVar, Union

from ..common import TypeHint, VarTuple
from ..feature_requirement import HAS_ANNOTATED, HAS_PY_39, HAS_PY_312, HAS_STD_CLASSES_GENERICS
from .constants import BUILTIN_ORIGIN_TO_TYPEVARS
from .fundamentals import get_generic_args, get_type_vars, strip_alias


def is_subclass_soft(cls, classinfo) -> bool:
    """Acts like builtin issubclass, but returns False instead of rising TypeError"""
    try:
        return issubclass(cls, classinfo)
    except TypeError:
        return False


_NEW_TYPE_CLS = type(NewType("", None))


def is_new_type(tp) -> bool:
    return isinstance(tp, _NEW_TYPE_CLS)


_TYPED_DICT_MCS = type(types.new_class("_TypedDictSample", (TypedDict,), {}))


def is_typed_dict_class(tp) -> bool:
    return isinstance(tp, _TYPED_DICT_MCS)


_NAMED_TUPLE_METHODS = ("_fields", "_field_defaults", "_make", "_replace", "_asdict")


def is_named_tuple_class(tp) -> bool:
    return (
        is_subclass_soft(tp, tuple)
        and all(
            hasattr(tp, attr_name)
            for attr_name in _NAMED_TUPLE_METHODS
        )
    )


def is_protocol(tp):
    if not isinstance(tp, type):
        return False

    return Protocol in tp.__bases__


def create_union(args: tuple):
    return Union[args]


def is_parametrized(tp: TypeHint) -> bool:
    return bool(get_generic_args(tp))


if HAS_PY_312:
    def is_user_defined_generic(tp: TypeHint) -> bool:
        return (
            bool(get_type_vars(tp))
            and (
                is_subclass_soft(strip_alias(tp), Generic)
                or isinstance(tp, typing.TypeAliasType)  # type: ignore[attr-defined]
            )
        )
else:
    def is_user_defined_generic(tp: TypeHint) -> bool:
        return (
            bool(get_type_vars(tp))
            and is_subclass_soft(strip_alias(tp), Generic)
        )


def is_generic(tp: TypeHint) -> bool:
    """Check if the type could be parameterized"""
    return (
        bool(get_type_vars(tp))
        or (
            strip_alias(tp) in BUILTIN_ORIGIN_TO_TYPEVARS
            and tp != type
            and not is_parametrized(tp)
            and (
                bool(HAS_STD_CLASSES_GENERICS) or not isinstance(tp, type)
            )
        )
        or (
            bool(HAS_ANNOTATED)
            and strip_alias(tp) == typing.Annotated
            and tp != typing.Annotated
            and is_generic(tp.__origin__)
        )
    )


def is_bare_generic(tp: TypeHint) -> bool:
    """Check if the type could be parameterized, excluding type aliases (list[T] etc.)"""
    return (
        (
            is_generic(strip_alias(tp))
            # for 3.8 and List (list is not generic)
            or is_generic(tp)
            # at 3.8 list is bare_generic but not generic
            # (this function only needs to create predicate)
            or tp in BUILTIN_ORIGIN_TO_TYPEVARS
        )
        and not is_parametrized(tp)
    )


def is_generic_class(cls: type) -> bool:
    """Check if the class represents a generic type.
    This function is faster than ``.is_generic()``, but it is limited to testing only classes
    """
    return (
        cls in BUILTIN_ORIGIN_TO_TYPEVARS
        or (
            issubclass(cls, Generic)  # type: ignore[arg-type]
            and bool(cls.__parameters__)  # type: ignore[attr-defined]
        )
    )


def get_type_vars_of_parametrized(tp: TypeHint) -> VarTuple[TypeVar]:
    params = get_type_vars(tp)
    if not params:
        return ()
    if isinstance(tp, type):
        if HAS_STD_CLASSES_GENERICS and isinstance(tp, types.GenericAlias):
            return params
        return ()
    if strip_alias(tp) != tp and get_generic_args(tp) == ():
        return ()
    return params


if HAS_PY_39:
    def eval_forward_ref(namespace: Dict[str, Any], forward_ref: ForwardRef):
        return forward_ref._evaluate(namespace, None, frozenset())
else:
    def eval_forward_ref(namespace: Dict[str, Any], forward_ref: ForwardRef):
        return forward_ref._evaluate(namespace, None)  # type: ignore[call-arg]



from .basic_utils import (
    create_union,
    is_bare_generic,
    is_generic,
    is_generic_class,
    is_named_tuple_class,
    is_new_type,
    is_parametrized,
    is_protocol,
    is_subclass_soft,
    is_typed_dict_class,
    is_user_defined_generic,
)
from .fundamentals import get_all_type_hints, get_generic_args, get_type_vars, is_pydantic_class, strip_alias
from .norm_utils import is_class_var, strip_tags
from .normalize_type import (
    AnyNormTypeVarLike,
    BaseNormType,
    NormParamSpecMarker,
    NormTV,
    NormTVTuple,
    NormTypeAlias,
    make_norm_type,
    normalize_type,
)



from __future__ import annotations

from typing import Any

from dishka.entities.component import DEFAULT_COMPONENT, Component
from dishka.entities.key import DependencyKey
from dishka.entities.scope import BaseScope
from .alias import Alias
from .factory import Factory, FactoryType


def context_stub() -> Any:
    raise NotImplementedError


class ContextVariable:
    __slots__ = ("provides", "scope")

    def __init__(
            self, *,
            provides: DependencyKey,
            scope: BaseScope | None,
    ) -> None:
        self.provides = provides
        self.scope = scope

    def as_factory(
            self, component: Component,
    ) -> Factory:
        if component == DEFAULT_COMPONENT:
            return Factory(
                scope=self.scope,
                source=context_stub,
                provides=self.provides,
                is_to_bind=False,
                dependencies=[],
                kw_dependencies={},
                type_=FactoryType.CONTEXT,
                cache=False,
            )
        else:
            aliased = Alias(
                source=self.provides.with_component(DEFAULT_COMPONENT),
                provides=DependencyKey(self.provides.type_hint,
                                       component=component),
                cache=False,
            )
            return aliased.as_factory(scope=self.scope, component=component)

    def __get__(self, instance: Any, owner: Any) -> ContextVariable:
        scope = self.scope or instance.scope
        return ContextVariable(
            scope=scope,
            provides=self.provides,
        )



__all__ = [
    "FromDishka",
    "inject",
    "DishkaAioInterceptor",
    "DishkaAioInterceptor",
]

from collections.abc import Awaitable, Callable, Iterator
from contextvars import ContextVar
from inspect import isasyncgenfunction, iscoroutinefunction
from typing import Any, ParamSpec, TypeVar

import grpc
from google.protobuf import message

from dishka import AsyncContainer, Container, FromDishka, Scope
from dishka.integrations.base import wrap_injection

P = ParamSpec("P")
RT = TypeVar("RT")

_dishka_scoped_container = ContextVar("_dishka_scoped_container")


def inject(func: Callable[P, RT]) -> Callable[P, RT]:
    return wrap_injection(
        func=func,
        is_async=iscoroutinefunction(func) or isasyncgenfunction(func),
        container_getter=lambda _, __: _dishka_scoped_container.get(),
    )


class DishkaInterceptor(grpc.ServerInterceptor):  # type: ignore[misc]
    def __init__(self, container: Container) -> None:
        self._container = container

    def intercept_service(
        self,
        continuation: Callable[
            [grpc.HandlerCallDetails],
            grpc.RpcMethodHandler,
        ],
        handler_call_details: grpc.HandlerCallDetails,
    ) -> grpc.RpcMethodHandler:
        rpc_handler = continuation(handler_call_details)

        def unary_unary_behavior(
            request: message.Message,
            context: grpc.ServicerContext,
        ) -> Any:
            context_ = {
                message.Message: request,
                grpc.ServicerContext: context,
            }
            with self._container(context=context_) as container:
                _dishka_scoped_container.set(container)
                return rpc_handler.unary_unary(request, context)

        def stream_unary_behavior(
            request_iterator: Iterator[message.Message],
            context: grpc.ServicerContext,
        ) -> Any:
            context_ = {grpc.ServicerContext: context}
            with self._container(
                context=context_,
                scope=Scope.SESSION,
            ) as container:
                _dishka_scoped_container.set(container)
                return rpc_handler.stream_unary(
                    request_iterator, context,
                )

        def unary_stream_behavior(
            request: message.Message,
            context: grpc.ServicerContext,
        ) -> Any:
            context_ = {
                message.Message: request,
                grpc.ServicerContext: context,
            }
            with self._container(context=context_) as container:
                _dishka_scoped_container.set(container)
                yield from rpc_handler.unary_stream(request, context)

        def stream_stream_behavior(
            request_iterator: Iterator[message.Message],
            context: grpc.ServicerContext,
        ) -> Any:
            context_ = {grpc.ServicerContext: context}
            with self._container(
                context=context_,
                scope=Scope.SESSION,
            ) as container:
                _dishka_scoped_container.set(container)
                yield from rpc_handler.stream_stream(request_iterator, context)

        if rpc_handler.unary_unary:
            return grpc.unary_unary_rpc_method_handler(
                unary_unary_behavior,
                rpc_handler.request_deserializer,
                rpc_handler.response_serializer,
            )
        elif rpc_handler.stream_unary:
            return grpc.stream_unary_rpc_method_handler(
                stream_unary_behavior,
                rpc_handler.request_deserializer,
                rpc_handler.response_serializer,
            )
        elif rpc_handler.unary_stream:
            return grpc.unary_stream_rpc_method_handler(
                unary_stream_behavior,
                rpc_handler.request_deserializer,
                rpc_handler.response_serializer,
            )
        elif rpc_handler.stream_stream:
            return grpc.stream_stream_rpc_method_handler(
                stream_stream_behavior,
                rpc_handler.request_deserializer,
                rpc_handler.response_serializer,
            )

        return rpc_handler


class DishkaAioInterceptor(grpc.aio.ServerInterceptor):  # type: ignore[misc]
    def __init__(self, container: AsyncContainer) -> None:
        self._container = container

    async def intercept_service(  # noqa: C901
        self,
        continuation: Callable[
            [grpc.HandlerCallDetails],
            Awaitable[grpc.RpcMethodHandler],
        ],
        handler_call_details: grpc.HandlerCallDetails,
    ) -> grpc.RpcMethodHandler:
        rpc_handler = await continuation(handler_call_details)

        async def unary_unary_behavior(
            request: message.Message,
            context: grpc.ServicerContext,
        ) -> Any:
            context_ = {
                message.Message: request,
                grpc.ServicerContext: context,
            }
            async with self._container(context=context_) as container:
                _dishka_scoped_container.set(container)
                return await rpc_handler.unary_unary(request, context)

        async def stream_unary_behavior(
            request_iterator: Iterator[message.Message],
            context: grpc.ServicerContext,
        ) -> Any:
            context_ = {grpc.ServicerContext: context}
            async with self._container(
                context=context_,
                scope=Scope.SESSION,
            ) as container:
                _dishka_scoped_container.set(container)
                return await rpc_handler.stream_unary(
                    request_iterator, context,
                )

        async def unary_stream_behavior(
            request: message.Message,
            context: grpc.ServicerContext,
        ) -> Any:
            context_ = {
                message.Message: request,
                grpc.ServicerContext: context,
            }
            async with self._container(
                context=context_,
                scope=Scope.REQUEST,
            ) as container:
                _dishka_scoped_container.set(container)
                stream = rpc_handler.unary_stream(request, context)
                async for result in stream:
                    yield result

        async def stream_stream_behavior(
            request_iterator: Iterator[message.Message],
            context: grpc.ServicerContext,
        ) -> Any:
            context_ = {grpc.ServicerContext: context}
            async with self._container(
                context=context_,
                scope=Scope.SESSION,
            ) as container:
                _dishka_scoped_container.set(container)
                stream = rpc_handler.stream_stream(request_iterator, context)
                async for result in stream:
                    yield result

        if rpc_handler.unary_unary:
            return grpc.unary_unary_rpc_method_handler(
                unary_unary_behavior,
                rpc_handler.request_deserializer,
                rpc_handler.response_serializer,
            )
        elif rpc_handler.stream_unary:
            return grpc.stream_unary_rpc_method_handler(
                stream_unary_behavior,
                rpc_handler.request_deserializer,
                rpc_handler.response_serializer,
            )
        elif rpc_handler.unary_stream:
            return grpc.unary_stream_rpc_method_handler(
                unary_stream_behavior,
                rpc_handler.request_deserializer,
                rpc_handler.response_serializer,
            )
        elif rpc_handler.stream_stream:
            return grpc.stream_stream_rpc_method_handler(
                stream_stream_behavior,
                rpc_handler.request_deserializer,
                rpc_handler.response_serializer,
            )

        return rpc_handler



from collections.abc import Awaitable, Callable, Sequence
from inspect import (
    Parameter,
    Signature,
    isasyncgenfunction,
    isgeneratorfunction,
    signature,
)
from typing import (
    Annotated,
    Any,
    Literal,
    ParamSpec,
    TypeAlias,
    TypeVar,
    cast,
    get_args,
    get_origin,
    get_type_hints,
    overload,
)

from dishka.async_container import AsyncContainer
from dishka.container import Container
from dishka.entities.component import DEFAULT_COMPONENT
from dishka.entities.depends_marker import FromDishka
from dishka.entities.key import DependencyKey, FromComponent

T = TypeVar("T")
P = ParamSpec("P")
DependencyParser: TypeAlias = Callable[[Parameter, Any], DependencyKey | None]
ContainerGetter: TypeAlias = Callable[[tuple[Any, ...], dict[str, Any]], T]
DependsClass: TypeAlias = cast(
    type | Sequence[type],
    FromDishka | FromComponent,
)


def default_parse_dependency(
        parameter: Parameter,
        hint: Any,
) -> DependencyKey | None:
    """Resolve dependency type or return None if it is not a dependency."""
    if get_origin(hint) is not Annotated:
        return None
    args = get_args(hint)
    dep = next(
        (arg for arg in args if isinstance(arg, DependsClass)),
        None,
    )
    if not dep:
        return None
    if isinstance(dep, FromDishka | FromComponent):  # type: ignore[arg-type]
        return DependencyKey(args[0], dep.component)
    else:
        return DependencyKey(args[0], DEFAULT_COMPONENT)


@overload
def wrap_injection(
        *,
        func: Callable[P, T],
        container_getter: ContainerGetter[Container],
        is_async: Literal[False] = False,
        remove_depends: bool = True,
        additional_params: Sequence[Parameter] = (),
        parse_dependency: DependencyParser = default_parse_dependency,
) -> Callable[P, T]:
    ...


@overload
def wrap_injection(
        *,
        func: Callable[P, T],
        container_getter: ContainerGetter[AsyncContainer],
        is_async: Literal[True] = True,
        remove_depends: bool = True,
        additional_params: Sequence[Parameter] = (),
        parse_dependency: DependencyParser = default_parse_dependency,
) -> Callable[P, T]:
    ...


def wrap_injection(
        *,
        func: Callable[P, T],
        container_getter: ContainerGetter[Container | AsyncContainer],
        is_async: bool = False,
        remove_depends: bool = True,
        additional_params: Sequence[Parameter] = (),
        parse_dependency: DependencyParser = default_parse_dependency,
) -> Callable[P, T]:
    hints = get_type_hints(func, include_extras=True)
    func_signature = signature(func)

    dependencies = {}
    for name, param in func_signature.parameters.items():
        hint = hints.get(name, Any)
        dep = parse_dependency(param, hint)
        if dep is None:
            continue
        dependencies[name] = dep

    if remove_depends:
        new_annotations = {
            name: hint
            for name, hint in hints.items()
            if name not in dependencies
        }
        new_params = [
            param
            for name, param in func_signature.parameters.items()
            if name not in dependencies
        ]
    else:
        new_annotations = hints.copy()
        new_params = list(func_signature.parameters.copy().values())

    auto_injected_func: Callable[P, T | Awaitable[T]]
    if additional_params:
        new_params.extend(additional_params)
        for param in additional_params:
            new_annotations[param.name] = param.annotation

    if is_async:
        auto_injected_func = _async_injection_wrapper(
            # typing.cast is needed because the function must be async
            func=cast(Callable[P, Awaitable[T]],func),
            dependencies=dependencies,
            additional_params=additional_params,
            container_getter=cast(
                ContainerGetter[AsyncContainer], container_getter,
            ),
        )
    else:
        auto_injected_func = _sync_injection_wrapper(
            func=func,
            dependencies=dependencies,
            additional_params=additional_params,
            container_getter=cast(
                ContainerGetter[Container], container_getter,
            ),
        )

    auto_injected_func.__dishka_injected__ = True  # type: ignore[attr-defined]
    auto_injected_func.__name__ = func.__name__
    auto_injected_func.__qualname__ = func.__qualname__
    auto_injected_func.__doc__ = func.__doc__
    auto_injected_func.__annotations__ = new_annotations
    auto_injected_func.__signature__ = Signature(  # type: ignore[attr-defined]
        parameters=new_params,
        return_annotation=func_signature.return_annotation,
    )
    return cast(Callable[P, T], auto_injected_func)


def is_dishka_injected(func: Callable[..., Any]) -> bool:
    return hasattr(func, "__dishka_injected__")


def _async_injection_wrapper(
        container_getter: ContainerGetter[AsyncContainer],
        additional_params: Sequence[Parameter],
        dependencies: dict[str, DependencyKey],
        func: Callable[P, Awaitable[T]],
) -> Callable[P, Awaitable[T]]:
    if isasyncgenfunction(func):
        async def auto_injected_func(*args: P.args, **kwargs: P.kwargs) -> T:
            container = container_getter(args, kwargs)
            for param in additional_params:
                kwargs.pop(param.name)
            solved = {
                name: await container.get(
                    dep.type_hint, component=dep.component,
                )
                for name, dep in dependencies.items()
            }
            async for message in func(*args, **kwargs, **solved):
                yield message
    else:
        async def auto_injected_func(*args: P.args, **kwargs: P.kwargs) -> T:
            container = container_getter(args, kwargs)
            for param in additional_params:
                kwargs.pop(param.name)
            solved = {
                name: await container.get(
                    dep.type_hint, component=dep.component,
                )
                for name, dep in dependencies.items()
            }
            return await func(*args, **kwargs, **solved)

    return auto_injected_func


def _sync_injection_wrapper(
        container_getter: ContainerGetter[Container],
        additional_params: Sequence[Parameter],
        dependencies: dict[str, DependencyKey],
        func: Callable[P, T],
) -> Callable[P, T]:
    if isgeneratorfunction(func):
        def auto_injected_func(*args: P.args, **kwargs: P.kwargs) -> T:
            container = container_getter(args, kwargs)
            for param in additional_params:
                kwargs.pop(param.name)
            solved = {
                name: container.get(dep.type_hint, component=dep.component)
                for name, dep in dependencies.items()
            }
            yield from func(*args, **kwargs, **solved)
    else:
        def auto_injected_func(*args: P.args, **kwargs: P.kwargs) -> T:
            container = container_getter(args, kwargs)
            for param in additional_params:
                kwargs.pop(param.name)
            solved = {
                name: container.get(dep.type_hint, component=dep.component)
                for name, dep in dependencies.items()
            }
            return func(*args, **kwargs, **solved)

    return auto_injected_func



from collections.abc import Sequence

from dishka.dependency_source import Factory, FactoryType
from dishka.entities.component import Component
from dishka.entities.key import DependencyKey
from dishka.entities.scope import BaseScope


class PathRenderer:

    def _arrow_cycle(self, index: int, length: int) -> str:
        if length == 1:
            return "⥁"
        elif index == 0:
            return " → → "
        elif index + 1 < length:
            return "↑   ↓"
        else:
            return " ← ← "

    def _arrow_line(self, index: int, length: int) -> str:
        if index + 1 < length:
            return "↓ "
        else:
            return " →"

    def _key(self, key: DependencyKey) -> str:
        return str(key.type_hint)

    def _source(self, factory: Factory) -> str:
        source = factory.source
        if source == factory.provides.type_hint:
            return ""
        if factory.type is FactoryType.ALIAS:
            return "alias"
        if func := getattr(source, "__func__", None):
            return getattr(func, "__qualname__", None) or str(func)
        else:
            return str(source)

    def _switch(
            self, scope: BaseScope | None, component: Component | None,
    ) -> str:
        return f"~~~ component={component!r}, {scope} ~~~\n"

    def render(
            self,
            path: Sequence[Factory],
            last: DependencyKey | None = None,
    ) -> str:
        if last is None:
            _arrow = self._arrow_cycle
        else:
            _arrow = self._arrow_line

        width = max(len(self._key(x.provides)) for x in path)
        if last:
            width = max(width, len(self._key(last)))

        dest: tuple[BaseScope | None, Component | None] = (None, "")
        length = len(path) + bool(last)

        res = ""
        for i, factory in enumerate(path):
            arrow = _arrow(i, length)
            new_dest = (factory.scope, factory.provides.component)
            if new_dest != dest:
                res += "   " + " " * len(arrow) + " " + self._switch(*new_dest)
                dest = new_dest

            res += (
                    "   " + arrow + " " +
                    self._key(factory.provides).ljust(width) +
                    " " +
                    self._source(factory) +
                    "\n"
            )
        if last:
            new_dest = (dest[0], last.component)
            arrow = _arrow(length + 1, length)
            if new_dest != dest:
                res += "   " + " " * len(arrow) + " " + self._switch(*new_dest)
                dest = new_dest
            res += (
                    "   " + arrow + " " +
                    self._key(last).ljust(width) +
                    " ???\n"
            )
        return res



import importlib.metadata
import itertools
import re
import sys
from abc import ABC, abstractmethod
from typing import Any, Iterable

from .common import VarTuple


def _true():
    return True


def _false():
    return False


class Requirement(ABC):
    __slots__ = ("is_meet", "__bool__", "__dict__")

    def __init__(self):
        self.is_meet = self._evaluate()
        self.__bool__ = _true if self.is_meet else _false

    @abstractmethod
    def _evaluate(self) -> bool:
        ...

    @property
    @abstractmethod
    def fail_reason(self) -> str:
        ...


class PythonVersionRequirement(Requirement):
    def __init__(self, min_version: VarTuple[int]):
        self.min_version = min_version
        super().__init__()

    def _evaluate(self) -> bool:
        return sys.version_info >= self.min_version

    @property
    def fail_reason(self) -> str:
        return f'Python >= {".".join(map(str, self.min_version))} is required'


class DistributionRequirement(Requirement):
    def __init__(self, distribution_name: str):
        self.distribution_name = distribution_name
        super().__init__()

    def _evaluate(self) -> bool:
        try:
            importlib.metadata.distribution(self.distribution_name)
        except importlib.metadata.PackageNotFoundError:
            return False
        return True

    @property
    def fail_reason(self) -> str:
        return f"Installed distribution {self.distribution_name!r} is required"


class DistributionVersionRequirement(DistributionRequirement):
    # Pattern from PEP 440
    _VERSION_PATTERN = r"""
        v?
        (?:
            (?:(?P<epoch>[0-9]+)!)?                           # epoch
            (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
            (?P<pre>                                          # pre-release
                [-_\.]?
                (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
                [-_\.]?
                (?P<pre_n>[0-9]+)?
            )?
            (?P<post>                                         # post release
                (?:-(?P<post_n1>[0-9]+))
                |
                (?:
                    [-_\.]?
                    (?P<post_l>post|rev|r)
                    [-_\.]?
                    (?P<post_n2>[0-9]+)?
                )
            )?
            (?P<dev>                                          # dev release
                [-_\.]?
                (?P<dev_l>dev)
                [-_\.]?
                (?P<dev_n>[0-9]+)?
            )?
        )
        (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
    """

    _version_regex = re.compile(
        r"^\s*" + _VERSION_PATTERN + r"\s*$",
        re.VERBOSE | re.IGNORECASE,
    )

    def __init__(self, distribution_name: str, min_version: str):
        self.min_version = min_version
        super().__init__(distribution_name)

    def _evaluate(self) -> bool:
        try:
            distribution = importlib.metadata.distribution(self.distribution_name)
        except importlib.metadata.PackageNotFoundError:
            return False
        return (
            self._make_comparator(distribution.version)
            >=
            self._make_comparator(self.min_version)
        )

    def _remove_trailing_zeros(self, data: Iterable[int]) -> VarTuple[int]:
        return tuple(reversed(tuple(itertools.dropwhile(lambda x: x == 0, reversed(tuple(data))))))

    def _make_comparator(self, version: str) -> VarTuple[Any]:
        match = self._version_regex.match(version)
        if match is None:
            raise ValueError
        return (
            match.group("epoch") or 0,
            self._remove_trailing_zeros(int(i) for i in match.group("release").split(".")),
            match.group("pre") is not None,
            match.group("dev") is not None,
        )

    @property
    def fail_reason(self) -> str:
        return f"Installed distribution {self.distribution_name!r} of {self.min_version!r} is required"


class PythonImplementationRequirement(Requirement):
    def __init__(self, implementation_name: str):
        self.implementation_name = implementation_name
        super().__init__()

    def _evaluate(self) -> bool:
        return sys.implementation.name == self.implementation_name

    @property
    def fail_reason(self) -> str:
        return f"{self.implementation_name} is required"


HAS_PY_39 = PythonVersionRequirement((3, 9))
HAS_ANNOTATED = HAS_PY_39
HAS_STD_CLASSES_GENERICS = HAS_PY_39

HAS_PY_310 = PythonVersionRequirement((3, 10))
HAS_TYPE_UNION_OP = HAS_PY_310
HAS_TYPE_GUARD = HAS_PY_310
HAS_TYPE_ALIAS = HAS_PY_310
HAS_PARAM_SPEC = HAS_PY_310

HAS_PY_311 = PythonVersionRequirement((3, 11))
HAS_NATIVE_EXC_GROUP = HAS_PY_311
HAS_TYPED_DICT_REQUIRED = HAS_PY_311
HAS_SELF_TYPE = HAS_PY_311
HAS_TV_TUPLE = HAS_PY_311
HAS_UNPACK = HAS_PY_311

HAS_PY_312 = PythonVersionRequirement((3, 12))
HAS_TV_SYNTAX = HAS_PY_312

HAS_SUPPORTED_ATTRS_PKG = DistributionVersionRequirement("attrs", "21.3.0")
HAS_ATTRS_PKG = DistributionRequirement("attrs")

HAS_SUPPORTED_SQLALCHEMY_PKG = DistributionVersionRequirement("sqlalchemy", "2.0.0")
HAS_SQLALCHEMY_PKG = DistributionRequirement("sqlalchemy")

HAS_SUPPORTED_PYDANTIC_PKG = DistributionVersionRequirement("pydantic", "2.0.0")
HAS_PYDANTIC_PKG = DistributionRequirement("pydantic")

IS_CPYTHON = PythonImplementationRequirement("cpython")
IS_PYPY = PythonImplementationRequirement("pypy")



__all__ = [
    "inject",
    "setup_dishka",
]

from collections.abc import Awaitable, Callable
from typing import Any, Final, ParamSpec, TypeVar

from arq import Worker
from arq.typing import StartupShutdown

from dishka.async_container import AsyncContainer
from dishka.integrations.base import wrap_injection

T = TypeVar("T")
P = ParamSpec("P")
DISHKA_APP_CONTAINER_KEY: Final = "dishka_app_container"
DISHKA_REQUEST_CONTAINER_KEY: Final = "dishka_request_container"


def inject(func: Callable[P, T]) -> Callable[P, T]:
    return wrap_injection(
        func=func,
        remove_depends=True,
        container_getter=lambda p, _: p[0][DISHKA_REQUEST_CONTAINER_KEY],
        is_async=True,
    )


def job_start(
    hook_func: StartupShutdown | None,
) -> Callable[[dict[Any, Any]], Awaitable[None]]:
    async def wrapper(context: dict[Any, Any]) -> None:
        container: AsyncContainer = context[DISHKA_APP_CONTAINER_KEY]
        sub_container = await container().__aenter__()
        context[DISHKA_REQUEST_CONTAINER_KEY] = sub_container
        if hook_func:
            await hook_func(context)

    return wrapper


def job_end(
    hook_func: StartupShutdown | None,
) -> Callable[[dict[Any, Any]], Awaitable[None]]:
    async def wrapper(context: dict[Any, Any]) -> None:
        if hook_func:
            await hook_func(context)

        sub_container: AsyncContainer = context[DISHKA_REQUEST_CONTAINER_KEY]
        await sub_container.close()

    return wrapper


def setup_dishka(
    container: AsyncContainer,
    worker_settings: dict[Any, Any] | Worker | Any,
) -> None:
    if isinstance(worker_settings, dict):
        if worker_settings.get("ctx"):
            worker_settings["ctx"][DISHKA_APP_CONTAINER_KEY] = container
        else:
            worker_settings["ctx"] = {DISHKA_APP_CONTAINER_KEY: container}

        worker_settings["on_job_start"] = job_start(
            worker_settings.get("on_job_start"),
        )
        worker_settings["on_job_end"] = job_end(
            worker_settings.get("on_job_end"),
        )
    else:
        if hasattr(worker_settings, "ctx"):
            worker_settings.ctx[DISHKA_APP_CONTAINER_KEY] = container
        else:
            worker_settings.ctx = {DISHKA_APP_CONTAINER_KEY: container}

        worker_settings.on_job_start = (
            job_start(worker_settings.on_job_start)
            if hasattr(worker_settings, "on_job_start")
            else job_start(None)
        )
        worker_settings.on_job_end = (
            job_end(worker_settings.on_job_end)
            if hasattr(worker_settings, "on_job_end")
            else job_end(None)
        )



from asyncio import iscoroutinefunction
from collections.abc import (
    AsyncGenerator,
    AsyncIterable,
    AsyncIterator,
    Callable,
    Generator,
    Iterable,
    Iterator,
    Sequence,
)
from inspect import (
    Parameter,
    isasyncgenfunction,
    isbuiltin,
    isclass,
    isfunction,
    isgeneratorfunction,
    ismethod,
    signature,
    unwrap,
)
from typing import (
    Annotated,
    Any,
    Protocol,
    TypeAlias,
    cast,
    get_args,
    get_origin,
    get_type_hints,
    overload,
)

from dishka._adaptix.type_tools.basic_utils import (  # type: ignore[attr-defined]
    get_type_vars,
    is_bare_generic,
    strip_alias,
)
from dishka._adaptix.type_tools.fundamentals import (
    get_all_type_hints,
)
from dishka._adaptix.type_tools.generic_resolver import (
    GenericResolver,
    MembersStorage,
)
from dishka.entities.key import (
    hint_to_dependency_key,
    hints_to_dependency_keys,
)
from dishka.entities.scope import BaseScope
from .composite import CompositeDependencySource, ensure_composite
from .factory import Factory, FactoryType
from .unpack_provides import unpack_factory

_empty = signature(lambda a: 0).parameters["a"].annotation
_protocol_init = type("_stub_proto", (Protocol,), {}).__init__  # type: ignore[misc, arg-type]
ProvideSource: TypeAlias = (
    Callable[..., Any]
    | classmethod  # type: ignore[type-arg]
    | staticmethod  # type: ignore[type-arg]
    | type
)


def _is_bound_method(obj: Any) -> bool:
    return ismethod(obj) and bool(obj.__self__)


def _get_init_members(tp: type) -> MembersStorage[str, None]:
    type_hints = get_all_type_hints(tp.__init__)  # type: ignore[misc, no-untyped-call]
    if "__init__" in tp.__dict__:
        overridden = frozenset(type_hints)
    else:
        overridden = frozenset()

    return MembersStorage(
        meta=None,
        members=type_hints,
        overriden=overridden,
    )


def _guess_factory_type(source: Any) -> FactoryType:
    if isasyncgenfunction(source):
        return FactoryType.ASYNC_GENERATOR
    elif isgeneratorfunction(source):
        return FactoryType.GENERATOR
    elif iscoroutinefunction(source):
        return FactoryType.ASYNC_FACTORY
    else:
        return FactoryType.FACTORY


def _type_repr(hint: Any) -> str:
    if hint is type(None):
        return "None"
    module = getattr(hint, "__module__", "")
    if module == "builtins":
        module = ""
    elif module:
        module += "."
    try:
        return f"{module}{hint.__qualname__}"
    except AttributeError:
        return str(hint)


def _async_generator_result(hint: Any) -> Any:
    origin = get_origin(hint)
    if origin is AsyncIterable:
        return get_args(hint)[0]
    elif origin is AsyncIterator:
        return get_args(hint)[0]
    elif origin is AsyncGenerator:
        return get_args(hint)[0]
    # errors
    name = _type_repr(hint)
    if origin is Iterable:
        args = ", ".join(_type_repr(a) for a in get_args(hint))
        guess = "AsyncIterable"
    elif origin is Iterator:
        args = ", ".join(_type_repr(a) for a in get_args(hint))
        guess = "AsyncIterator"
    elif origin is Generator:
        args = ", ".join(_type_repr(a) for a in get_args(hint)[:2])
        guess = "AsyncGenerator"
    else:
        args = name
        guess = "AsyncIterable"

    raise TypeError(
        f"Unsupported return type `{name}` for async generator. "
        f"Did you mean {guess}[{args}]?",
    )


def _generator_result(hint: Any) -> Any:
    origin = get_origin(hint)
    if origin is Iterable:
        return get_args(hint)[0]
    elif origin is Iterator:
        return get_args(hint)[0]
    elif origin is Generator:
        return get_args(hint)[1]
    # errors
    name = _type_repr(hint)
    if origin is AsyncIterable:
        args = ", ".join(_type_repr(a) for a in get_args(hint))
        guess = "Iterable"
    elif origin is AsyncIterator:
        args = ", ".join(_type_repr(a) for a in get_args(hint))
        guess = "Iterator"
    elif origin is AsyncGenerator:
        args = ", ".join(_type_repr(a) for a in get_args(hint)) + ", None"
        guess = "Generator"
    else:
        args = name
        guess = "Iterable"

    raise TypeError(
        f"Unsupported return type `{name}` for generator. "
        f"Did you mean {guess}[{args}]?",
    )


def _clean_result_hint(
    factory_type: FactoryType,
    possible_dependency: Any,
) -> Any:
    if factory_type == FactoryType.ASYNC_GENERATOR:
        return _async_generator_result(possible_dependency)
    elif factory_type == FactoryType.GENERATOR:
        return _generator_result(possible_dependency)
    return possible_dependency


def _params_without_hints(func: Any, *, skip_self: bool) -> Sequence[str]:
    if func is object.__init__:
        return []
    if func is _protocol_init:
        return []
    params = signature(func).parameters
    return [
        p.name
        for i, p in enumerate(params.values())
        if p.annotation is _empty
        if i > 0 or not skip_self
    ]


def _make_factory_by_class(
        *,
        provides: Any,
        scope: BaseScope | None,
        source: type,
        cache: bool,
) -> Factory:
    if not provides:
        provides = source

    if get_origin(source) is Annotated:
        source = get_args(source)[0]
    init = strip_alias(source).__init__
    if missing_hints := _params_without_hints(init, skip_self=True):
        name = f"{source.__module__}.{source.__qualname__}.__init__"
        missing = ", ".join(missing_hints)
        raise ValueError(
            f"Failed to analyze `{name}`. \n"
            f"Some parameters do not have type hints: {missing}\n",
        )
    # we need to fix concrete generics and normal classes as well
    # as classes can be children of concrete generics
    res = GenericResolver(_get_init_members)
    try:
        hints = dict(res.get_resolved_members(source).members)
    except NameError as e:
        name = f"{source.__module__}.{source.__qualname__}.__init__"
        raise NameError(
            f"Failed to analyze `{name}`. \n"
            f"Type '{e.name}' is not defined\n\n"
            f"If your are using `if TYPE_CHECKING` to import '{e.name}' "
            f"then try removing it. \n"
            f"Or, create a separate factory with all types imported.",
            name=e.name,
        ) from e

    hints.pop("return", _empty)
    params = signature(init).parameters
    kw_dependency_keys = {
        name: hint_to_dependency_key(hints.pop(name))
        for name, param in params.items()
        if param.kind is Parameter.KEYWORD_ONLY
    }
    dependencies = list(hints.values())

    return Factory(
        dependencies=hints_to_dependency_keys(dependencies),
        kw_dependencies=kw_dependency_keys,
        type_=FactoryType.FACTORY,
        source=source,
        scope=scope,
        provides=hint_to_dependency_key(provides),
        is_to_bind=False,
        cache=cache,
    )


def _make_factory_by_function(
        *,
        provides: Any,
        scope: BaseScope | None,
        source: Callable[..., Any] | classmethod, # type: ignore[type-arg]
        cache: bool,
        is_in_class: bool,
) -> Factory:
    # typing.cast is applied as unwrap takes a Callable object
    raw_source = unwrap(cast(Callable[..., Any], source))
    missing_hints = _params_without_hints(raw_source, skip_self=is_in_class)
    if missing_hints:
        name = getattr(source, "__qualname__", "") or str(source)
        missing = ", ".join(missing_hints)
        raise ValueError(
            f"Failed to analyze `{name}`. \n"
            f"Some parameters do not have type hints: {missing}\n",
        )

    params = signature(raw_source).parameters
    factory_type = _guess_factory_type(raw_source)

    try:
        hints = get_type_hints(source, include_extras=True)
    except NameError as e:
        name = getattr(source, "__qualname__", "") or str(source)
        raise NameError(
            f"Failed to analyze `{name}`. \n"
            f"Type '{e.name}' is not defined. \n\n"
            f"If your are using `if TYPE_CHECKING` to import '{e.name}' "
            f"then try removing it. \n"
            f"Or, create a separate factory with all types imported.",
            name=e.name,
        ) from e
    if is_in_class:
        self = next(iter(params.values()), None)
        if self and self.name not in hints:
            # add self to dependencies, so it can be easily removed
            # if we will bind factory to provider instance
            hints = {self.name: Any, **hints}
    possible_dependency = hints.pop("return", _empty)

    kw_dependency_keys = {
        name: hint_to_dependency_key(hints.pop(name))
        for name, param in params.items()
        if param.kind is Parameter.KEYWORD_ONLY
    }
    dependencies = list(hints.values())

    if not provides:
        if possible_dependency is _empty:
            name = getattr(source, "__qualname__", "") or str(source)
            raise ValueError(f"Failed to analyze `{name}`. \n"
                             f"Missing return type hint.")
        try:
            provides = _clean_result_hint(factory_type, possible_dependency)
        except TypeError as e:
            name = getattr(source, "__qualname__", "") or str(source)
            raise TypeError(f"Failed to analyze `{name}`. \n" + str(e)) from e
    return Factory(
        dependencies=hints_to_dependency_keys(dependencies),
        kw_dependencies=kw_dependency_keys,
        type_=factory_type,
        source=source,
        scope=scope,
        provides=hint_to_dependency_key(provides),
        is_to_bind=is_in_class,
        cache=cache,
    )


def _make_factory_by_static_method(
        *,
        provides: Any,
        scope: BaseScope | None,
        source: staticmethod,  # type: ignore[type-arg]
        cache: bool,
) -> Factory:
    if missing_hints := _params_without_hints(source, skip_self=False):
        name = getattr(source, "__qualname__", "") or str(source)
        missing = ", ".join(missing_hints)
        raise ValueError(
            f"Failed to analyze `{name}`. \n"
            f"Some parameters do not have type hints: {missing}\n",
        )
    factory_type = _guess_factory_type(source.__wrapped__)
    try:
        hints = get_type_hints(source, include_extras=True)
    except NameError as e:
        name = getattr(source, "__qualname__", "") or str(source)
        raise NameError(
            f"Failed to analyze `{name}`. \n"
            f"Type '{e.name}' is not defined. \n\n"
            f"If your are using `if TYPE_CHECKING` to import '{e.name}' "
            f"then try removing it. \n"
            f"Or, create a separate factory with all types imported.",
            name=e.name,
        ) from e

    possible_dependency = hints.pop("return", _empty)

    params = signature(source).parameters
    kw_dependency_keys = {
        name: hint_to_dependency_key(hints.pop(name))
        for name, param in params.items()
        if param.kind is Parameter.KEYWORD_ONLY
    }
    dependencies = list(hints.values())

    if not provides:
        if possible_dependency is _empty:
            name = getattr(source, "__qualname__", "") or str(source)
            raise ValueError(f"Failed to analyze `{name}`. \n"
                             f"Missing return type hint.")
        try:
            provides = _clean_result_hint(factory_type, possible_dependency)
        except TypeError as e:
            name = getattr(source, "__qualname__", "") or str(source)
            raise TypeError(f"Failed to analyze `{name}`. \n" + str(e)) from e
    return Factory(
        dependencies=hints_to_dependency_keys(dependencies),
        kw_dependencies=kw_dependency_keys,
        type_=factory_type,
        source=source,
        scope=scope,
        provides=hint_to_dependency_key(provides),
        is_to_bind=False,
        cache=cache,
    )


def _make_factory_by_other_callable(
        *,
        provides: Any,
        scope: BaseScope | None,
        source: Callable[..., Any],
        cache: bool,
) -> Factory:
    if _is_bound_method(source):
        to_check = source.__func__  # type: ignore[attr-defined]
    else:
        to_check = type(source).__call__
    factory = make_factory(
        provides=provides,
        source=to_check,
        cache=cache,
        scope=scope,
        is_in_class=True,
    )
    if factory.is_to_bind:
        dependencies = factory.dependencies[1:]  # remove `self`
    else:
        dependencies = factory.dependencies
    return Factory(
        dependencies=dependencies,
        kw_dependencies=factory.kw_dependencies,
        type_=factory.type,
        source=source,
        scope=scope,
        provides=factory.provides,
        is_to_bind=False,
        cache=cache,
    )


def make_factory(
        *,
        provides: Any,
        scope: BaseScope | None,
        source: ProvideSource,
        cache: bool,
        is_in_class: bool,
) -> Factory:
    if is_bare_generic(source):
        source = source[get_type_vars(source)]  # type: ignore[index]

    if isclass(source) or get_origin(source):
        return _make_factory_by_class(
            provides=provides,
            scope=scope,
            source=cast(type, source),
            cache=cache,
        )
    elif isfunction(source) or isinstance(source, classmethod):
        return _make_factory_by_function(
            provides=provides,
            scope=scope,
            source=source,
            cache=cache,
            is_in_class=is_in_class,
        )
    elif isbuiltin(source):
        return _make_factory_by_function(
            provides=provides,
            scope=scope,
            source=source,
            cache=cache,
            is_in_class=False,
        )
    elif isinstance(source, staticmethod):
        return _make_factory_by_static_method(
            provides=provides,
            scope=scope,
            source=source,
            cache=cache,
        )
    elif callable(source):
        return _make_factory_by_other_callable(
            provides=provides,
            scope=scope,
            source=source,
            cache=cache,
        )
    else:
        raise TypeError(f"Cannot use {type(source)} as a factory")


def _provide(
        *,
        source: ProvideSource | None = None,
        scope: BaseScope | None = None,
        provides: Any = None,
        cache: bool = True,
        is_in_class: bool = True,
) -> CompositeDependencySource:
    composite = ensure_composite(source)
    factory = make_factory(
        provides=provides, scope=scope,
        source=composite.origin, cache=cache,
        is_in_class=is_in_class,
    )
    composite.dependency_sources.extend(unpack_factory(factory))
    return composite


def provide_on_instance(
        *,
        source: ProvideSource | None = None,
        scope: BaseScope | None = None,
        provides: Any = None,
        cache: bool = True,
) -> CompositeDependencySource:
    return _provide(
        provides=provides, scope=scope, source=source, cache=cache,
        is_in_class=False,
    )


@overload
def provide(
        *,
        scope: BaseScope | None = None,
        provides: Any = None,
        cache: bool = True,
) -> Callable[[Callable[..., Any]], CompositeDependencySource]:
    ...


@overload
def provide(
        source: ProvideSource | None,
        *,
        scope: BaseScope | None = None,
        provides: Any = None,
        cache: bool = True,
) -> CompositeDependencySource:
    ...


def provide(
        source: ProvideSource | None = None,
        *,
        scope: BaseScope | None = None,
        provides: Any = None,
        cache: bool = True,
) -> CompositeDependencySource | Callable[
    [Callable[..., Any]], CompositeDependencySource,
]:
    """
    Mark a method or class as providing some dependency.

    If used as a method decorator then return annotation is used
    to determine what is provided. User `provides` to override that.
    Method parameters are analyzed and passed automatically.

    If used with a class a first parameter than `__init__` method parameters
    are passed automatically. If no provides is passed then it is
    supposed that class itself is a provided dependency.

    Return value must be saved as a `Provider` class attribute and
    not intended for direct usage

    :param source: Method to decorate or class.
    :param scope: Scope of the dependency to limit its lifetime
    :param provides: Dependency type which is provided by this factory
    :return: instance of Factory or a decorator returning it
    :param cache: save created object to scope cache or not
    """
    if source is not None:
        return _provide(
            provides=provides, scope=scope, source=source, cache=cache,
            is_in_class=True,
        )

    def scoped(func: Callable[..., Any]) -> CompositeDependencySource:
        return _provide(
            provides=provides, scope=scope, source=func, cache=cache,
            is_in_class=True,
        )

    return scoped


def _provide_all(
        *,
        provides: Sequence[Any],
        scope: BaseScope | None,
        cache: bool,
        is_in_class: bool,
) -> CompositeDependencySource:
    composite = CompositeDependencySource(None)
    for single_provides in provides:
        factory = make_factory(
            source=single_provides,
            provides=single_provides,
            scope=scope,
            cache=cache,
            is_in_class=is_in_class,
        )
        composite.dependency_sources.extend(unpack_factory(factory))
    return composite


def provide_all(
        *provides: Any,
        scope: BaseScope | None = None,
        cache: bool = True,
) -> CompositeDependencySource:
    return _provide_all(
        provides=provides, scope=scope,
        cache=cache, is_in_class=True,
    )


def provide_all_on_instance(
        *provides: Any,
        scope: BaseScope | None = None,
        cache: bool = True,
) -> CompositeDependencySource:
    return _provide_all(
        provides=provides, scope=scope,
        cache=cache, is_in_class=False,
    )



import typing
from dataclasses import InitVar
from typing import ClassVar, Final, TypeVar

from ..feature_requirement import HAS_ANNOTATED, HAS_TYPED_DICT_REQUIRED
from .normalize_type import BaseNormType

_TYPE_TAGS = [Final, ClassVar, InitVar]

if HAS_ANNOTATED:
    _TYPE_TAGS.append(typing.Annotated)

if HAS_TYPED_DICT_REQUIRED:
    _TYPE_TAGS.extend([typing.Required, typing.NotRequired])


def strip_tags(norm: BaseNormType) -> BaseNormType:
    """Removes type hints that do not represent a type
    and that only indicates metadata
    """
    if norm.origin in _TYPE_TAGS:
        return strip_tags(norm.args[0])
    return norm


N = TypeVar("N", bound=BaseNormType)


def strip_annotated(value: N) -> N:
    if HAS_ANNOTATED and isinstance(value, BaseNormType) and value.origin == typing.Annotated:
        return strip_annotated(value)
    return value


def is_class_var(norm: BaseNormType) -> bool:
    if norm.origin == ClassVar:
        return True
    if norm.origin in _TYPE_TAGS:
        return is_class_var(norm.args[0])
    return False



import sys
import typing
from typing import Any, ForwardRef, Tuple, TypeVar

from ..common import TypeHint, VarTuple
from ..feature_requirement import HAS_PARAM_SPEC, HAS_TV_TUPLE
from .basic_utils import create_union, eval_forward_ref, is_user_defined_generic, strip_alias
from .constants import BUILTIN_ORIGIN_TO_TYPEVARS


class ImplicitParamsGetter:
    def _process_limit_element(self, type_var: TypeVar, tp: TypeHint) -> TypeHint:
        if isinstance(tp, ForwardRef):
            return eval_forward_ref(vars(sys.modules[type_var.__module__]), tp)
        return tp

    def _process_type_var(self, type_var) -> TypeHint:
        if HAS_PARAM_SPEC and isinstance(type_var, typing.ParamSpec):
            return ...
        if HAS_TV_TUPLE and isinstance(type_var, typing.TypeVarTuple):
            return typing.Unpack[Tuple[Any, ...]]
        if type_var.__constraints__:
            return create_union(
                tuple(
                    self._process_limit_element(type_var, constraint)
                    for constraint in type_var.__constraints__
                ),
            )
        if type_var.__bound__ is None:
            return Any
        return self._process_limit_element(type_var, type_var.__bound__)

    def get_implicit_params(self, origin) -> VarTuple[TypeHint]:
        if is_user_defined_generic(origin):
            type_vars = origin.__parameters__
        else:
            type_vars = BUILTIN_ORIGIN_TO_TYPEVARS.get(origin, ())

        return tuple(
            self._process_type_var(type_var)
            for type_var in type_vars
        )


def fill_implicit_params(tp: TypeHint) -> TypeHint:
    params = ImplicitParamsGetter().get_implicit_params(strip_alias(tp))
    if params:
        return tp[params]
    raise ValueError(f"Can not derive implicit parameters for {tp}")



__all__ = [
    "FromDishka",
    "inject",
    "setup_dishka",
]

from collections.abc import Callable
from inspect import Parameter
from typing import Any, ParamSpec, TypeVar

import telebot  # type: ignore[import-untyped]
from telebot import BaseMiddleware, TeleBot

from dishka import Container, FromDishka
from .base import wrap_injection

CONTAINER_NAME = "dishka_container"

T = TypeVar("T")
P = ParamSpec("P")

def inject(func: Callable[P, T]) -> Callable[P, T]:
    additional_params = [Parameter(
        name=CONTAINER_NAME,
        annotation=Container,
        kind=Parameter.KEYWORD_ONLY,
    )]

    return wrap_injection(
        func=func,
        additional_params=additional_params,
        container_getter=lambda _, p: p[CONTAINER_NAME],
    )


class ContainerMiddleware(BaseMiddleware):  # type: ignore[misc]
    update_types = telebot.util.update_types

    def __init__(self, container: Container) -> None:
        super().__init__()
        self.container = container

    def pre_process(
        self,
        message: Any,
        data: dict[str, Any],
    ) -> None:
        dishka_container_wrapper = self.container({type(message): message})
        data[CONTAINER_NAME + "_wrapper"] = dishka_container_wrapper
        data[CONTAINER_NAME] = dishka_container_wrapper.__enter__()

    def post_process(
        self,
        message: Any,
        data: dict[str, Any],
        exception: Exception,
    ) -> None:
        data[CONTAINER_NAME + "_wrapper"].__exit__(None, None, None)


def setup_dishka(container: Container, bot: TeleBot) -> Container:
    middleware = ContainerMiddleware(container)
    bot.setup_middleware(middleware)
    return container



from __future__ import annotations

from typing import Any

from dishka.entities.component import Component
from dishka.entities.key import DependencyKey
from dishka.entities.scope import BaseScope
from .factory import Factory


class Decorator:
    __slots__ = ("provides", "factory")

    def __init__(
            self,
            factory: Factory,
            provides: DependencyKey | None = None,
    ) -> None:
        self.factory = factory
        if provides:
            self.provides = provides
        else:
            self.provides = factory.provides

    def as_factory(
            self, *,
            scope: BaseScope,
            new_dependency: DependencyKey,
            cache: bool,
            component: Component,
    ) -> Factory:
        return Factory(
            scope=scope,
            source=self.factory.source,
            provides=self.factory.provides.with_component(component),
            is_to_bind=self.factory.is_to_bind,
            dependencies=[
                (
                    new_dependency
                    if dep.type_hint == self.provides.type_hint
                    else dep
                ).with_component(component)
                for dep in self.factory.dependencies
            ],
            kw_dependencies={
                name: (
                    new_dependency
                    if dep.type_hint == self.provides.type_hint
                    else dep
                ).with_component(component)
                for name, dep in self.factory.kw_dependencies.items()
            },
            type_=self.factory.type,
            cache=cache,
        )

    def __get__(self, instance: Any, owner: Any) -> Decorator:
        return Decorator(self.factory.__get__(instance, owner))
